__all__ = [
    'to_disk',
    'from_disk',
    'to_json',
    'from_json',
    'to_mpf',
    'from_mpf',
]

from matrixprofile.io.__io import to_disk
from matrixprofile.io.__io import from_disk
from matrixprofile.io.__io import to_json
from matrixprofile.io.__io import from_json
from matrixprofile.io.protobuf.protobuf_utils import (
	to_mpf,
	from_mpf
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

"""Tests for `scrimp` package."""

import os
import pytest

import numpy as np

import matrixprofile
from matrixprofile.algorithms import scrimp

MODULE_PATH = matrixprofile.__path__[0]


def test_time_series_too_short_exception():
    with pytest.raises(ValueError) as excinfo:
        scrimp.scrimp_plus_plus([1, 2, 3, 4, 5], 4, 0.25)
        assert 'Time series is too short' in str(excinfo.value)


def test_window_size_minimum_exception():
    with pytest.raises(ValueError) as excinfo:
        scrimp.scrimp_plus_plus([1, 2, 3, 4, 5], 2, 0.25)
        assert 'Window size must be at least 4' in str(excinfo.value)


def test_invalid_step_size_negative():
    exc = 'step_size should be a float between 0 and 1.'
    with pytest.raises(ValueError) as excinfo:
        scrimp.scrimp_plus_plus([1, 2, 3, 4, 5], 2, -1)
        assert exc in str(excinfo.value)


def test_invalid_step_size_str():
    exc = 'step_size should be a float between 0 and 1.'
    with pytest.raises(ValueError) as excinfo:
        scrimp.scrimp_plus_plus([1, 2, 3, 4, 5], 2, 'a')
        assert exc in str(excinfo.value)


def test_invalid_step_size_greater():
    exc = 'step_size should be a float between 0 and 1.'
    with pytest.raises(ValueError) as excinfo:
        scrimp.scrimp_plus_plus([1, 2, 3, 4, 5], 2, 2)
        assert exc in str(excinfo.value)


def test_invalid_random_state_exception():
    exc = 'Invalid random_state value given.'
    with pytest.raises(ValueError) as excinfo:
        scrimp.scrimp_plus_plus([1, 2, 3, 4, 5], 2, random_state='adsf')
        assert exc in str(excinfo.value)


def test_scrimp_plus_plus():
    ts = np.array([0, 0, 1, 0, 0, 0, 1, 0])
    m = 4
    step_size = 0.25
    profile = scrimp.scrimp_plus_plus(ts, m, step_size=step_size, sample_pct=1.0)

    expected_mp = np.array([
        0,
        3.2660,
        3.2660,
        3.2660,
        0
    ])
    expected_mpidx = np.array([
        4,
        3,
        0,
        0,
        0,
    ])

    np.testing.assert_almost_equal(profile['mp'], expected_mp, decimal=4)
    np.testing.assert_equal(profile['pi'], expected_mpidx)

    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    m = 32
    step_size = 0.25
    profile = scrimp.scrimp_plus_plus(ts, m, step_size=step_size, sample_pct=1.0)
    expected_mp = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'scrimp.mp.txt'))
    expected_mpi = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'scrimp.mpi.txt')).astype('int') - 1

    np.testing.assert_almost_equal(profile['mp'], expected_mp)
    np.testing.assert_equal(profile['pi'], expected_mpi)



# -*- coding: utf-8 -*-

"""
This module consists of all code to implement the SCRIMP++ algorithm. SCRIMP++ 
is an anytime algorithm that computes the matrix profile for a given time 
series (ts) over a given window size (m).

This algorithm was originally created at the University of California 
Riverside. For further academic understanding, please review this paper:

Matrix Proﬁle XI: SCRIMP++: Time Series Motif Discovery at Interactive
Speed. Yan Zhu, Chin-Chia Michael Yeh, Zachary Zimmerman, Kaveh Kamgar
Eamonn Keogh, ICDM 2018.

https://www.cs.ucr.edu/~eamonn/SCRIMP_ICDM_camera_ready_updated.pdf
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import math

import warnings

import numpy as np

from matrixprofile import core


def calc_distance_profile(X, y, n, m, meanx, sigmax):
    """
    Computes the distance profile.

    Parameters
    ----------
    X : array_like
        The FFT transformed time series.
    y : array_like
        The query.
    n : int
        The length of the time series.
    m : int
        The window size.
    meanx : array_like
        The moving mean of the time series.
    sigmax : array_like
        The moving standard deviation of the time series.

    Returns
    -------
    array_like :
        The distance profile.

    """
    # reverse the query
    y = np.flip(y, 0)
   
    # make y same size as ts with zero fill
    y = np.concatenate([y, np.zeros(n-m)])

    # main trick of getting dot product in O(n log n) time
    Y = np.fft.fft(y)
    Z = X * Y
    z = np.fft.ifft(Z)

    # compute y stats in O(n)
    sumy = np.sum(y)
    sumy2 = np.sum(np.power(y, 2))
    meany = sumy / m
    sigmay2 = sumy2 / m - meany ** 2
    sigmay = np.sqrt(sigmay2)

    dist = (z[m - 1:n] - m * meanx * meany) / (sigmax * sigmay)
    dist = m - dist
    dist = np.real(2 * dist)

    return np.sqrt(np.absolute(dist))


def calc_dotproduct_idx(dotproduct, m, mp, idx, sigmax, idx_nn, meanx):
    dotproduct[idx] = (m - mp[idx] ** 2 / 2) * \
        sigmax[idx] * sigmax[idx_nn] + m * meanx[idx] * meanx[idx_nn]

    return dotproduct


def calc_dotproduct_end_idx(ts, dp, idx, m, endidx, idx_nn, idx_diff):
    tmp_a = ts[idx+m:endidx+m]
    tmp_b = ts[idx_nn+m:endidx+m+idx_diff]
    tmp_c = ts[idx:endidx]
    tmp_d = ts[idx_nn:endidx+idx_diff]
    tmp_f = tmp_a * tmp_b - tmp_c * tmp_d

    dp[idx+1:endidx+1] = dp[idx] + np.cumsum(tmp_f)

    return dp


def calc_refine_distance_end_idx(refine_distance, dp, idx, endidx, meanx, 
                                 sigmax, idx_nn, idx_diff, m):
    tmp_a = dp[idx+1:endidx+1]
    tmp_b = meanx[idx+1:endidx+1]
    tmp_c = meanx[idx_nn+1:endidx+idx_diff+1]
    tmp_d = sigmax[idx+1:endidx+1]
    tmp_e = sigmax[idx_nn+1:endidx+idx_diff+1]
    tmp_f = tmp_b * tmp_c
    tmp_g = tmp_d * tmp_e
    tmp_h = (m-(tmp_a - m * tmp_f) / (tmp_g))
    refine_distance[idx+1:endidx+1] = np.sqrt(np.abs(2 * tmp_h))    

    return refine_distance


def calc_dotproduct_begin_idx(ts, dp, beginidx, idx, idx_diff, m, 
                              idx_nn):
    indices = list(range(idx - 1, beginidx - 1, -1))    

    if not indices:
        return dp

    tmp_a = ts[indices]
    indices_b = list(range(idx_nn - 1, beginidx + idx_diff - 1, -1))
    tmp_b = ts[indices_b]
    indices_c = list(range(idx + m - 1, beginidx + m - 1, -1))
    tmp_c = ts[indices_c]
    indices_d = list(range(idx_nn - 1 + m, beginidx + idx_diff + m - 1, -1))
    tmp_d = ts[indices_d]

    dp[indices] = dp[idx] + \
        np.cumsum((tmp_a * tmp_b) - (tmp_c * tmp_d))

    return dp


def calc_refine_distance_begin_idx(refine_distance, dp, beginidx, idx, 
                                   idx_diff, idx_nn, sigmax, meanx, m):
    if not (beginidx < idx):
        return refine_distance

    tmp_a = dp[beginidx:idx]
    tmp_b = meanx[beginidx:idx]
    tmp_c = meanx[beginidx+idx_diff:idx_nn]
    tmp_d = sigmax[beginidx:idx]
    tmp_e = sigmax[beginidx+idx_diff:idx_nn]
    tmp_f = tmp_b * tmp_c
    tmp_g = tmp_d * tmp_e
    tmp_h = (m-(tmp_a - m * tmp_f) / (tmp_g))

    refine_distance[beginidx:idx] = np.sqrt(np.abs(2 * tmp_h))

    return refine_distance


def apply_update_positions(matrix_profile, mp_index, refine_distance, beginidx,
                           endidx, orig_index, idx_diff):
    tmp_a = refine_distance[beginidx:endidx+1]
    tmp_b = matrix_profile[beginidx:endidx+1]
    update_pos1 = np.argwhere(tmp_a < tmp_b).flatten()    

    if len(update_pos1) > 0:        
        update_pos1 = update_pos1 + beginidx
        matrix_profile[update_pos1] = refine_distance[update_pos1]
        mp_index[update_pos1] = orig_index[update_pos1] + idx_diff

    tmp_a = refine_distance[beginidx:endidx + 1]
    tmp_b = matrix_profile[beginidx + idx_diff:endidx + idx_diff + 1]
    update_pos2 = np.argwhere(tmp_a < tmp_b).flatten()

    if len(update_pos2) > 0:
        update_pos2 = update_pos2 + beginidx
        matrix_profile[update_pos2 + idx_diff] = refine_distance[update_pos2]
        mp_index[update_pos2 + idx_diff] = orig_index[update_pos2]

    return (matrix_profile, mp_index)


def compute_indices(profile_len, step_size, sample_pct):
    """
    Computes the indices used for profile index iteration based on the
    number of samples and step size.

    Parameters
    ----------
    profile_len : int
        The length of the profile to be computed.
    step_size : float
        step_size : float, default 0.25
        The sampling interval for the window. The paper suggest 0.25 is the
        most practical. It should be a float value between 0 and 1.
    sample_pct : float, default = 0.1 (10%)
        Number of samples to compute distances for in the MP.

    Returns
    -------
    array_like :
        The indices to compute.

    """
    compute_order = np.arange(0, profile_len, step=step_size)
    sample_size = int(np.ceil(len(compute_order) * sample_pct))
    samples = np.random.choice(compute_order, size=sample_size, replace=False)

    return samples


def prescrimp(ts, window_size, query=None, step_size=0.25, sample_pct=0.1,
                     random_state=None, n_jobs=1):
    """
    This is the PreScrimp algorithm from the SCRIMP++ paper. It is primarly
    used to compute the approximate matrix profile. In this case we use
    a sample percentage to mock "the anytime/approximate nature".

    Parameters
    ----------
    ts : np.ndarray
        The time series to compute the matrix profile for.
    window_size : int
        The window size.
    query : array_like
        Optionally, a query can be provided to perform a similarity join.
    step_size : float, default 0.25
        The sampling interval for the window. The paper suggest 0.25 is the
        most practical. It should be a float value between 0 and 1.
    sample_pct : float, default = 0.1 (10%)
        Number of samples to compute distances for in the MP.
    random_state : int, default None
        Set the random seed generator for reproducible results.
    n_jobs : int, Default = 1
        Number of cpu cores to use.

    Note
    ----
    The matrix profiles computed from prescrimp will always be the approximate
    solution.

    Returns
    -------
    dict : profile
        A MatrixProfile data structure.
        
        >>> {
        >>>    'mp': The matrix profile,
        >>>    'pi': The matrix profile 1NN indices,
        >>>    'rmp': The right matrix profile,
        >>>    'rpi': The right matrix profile 1NN indices,
        >>>    'lmp': The left matrix profile,
        >>>    'lpi': The left matrix profile 1NN indices,
        >>>    'metric': The distance metric computed for the mp,
        >>>    'w': The window size used to compute the matrix profile,
        >>>    'ez': The exclusion zone used,
        >>>    'join': Flag indicating if a similarity join was computed,
        >>>    'sample_pct': Percentage of samples used in computing the MP,
        >>>    'data': {
        >>>        'ts': Time series data,
        >>>        'query': Query data if supplied
        >>>    }
        >>>    'class': "MatrixProfile"
        >>>    'algorithm': "prescrimp"
        >>>}

    Raises
    ------
    ValueError
        If window_size < 4.
        If window_size > query length / 2.
        If ts is not a list or np.array.
        If query is not a list or np.array.
        If ts or query is not one dimensional.
        If sample_pct is not between 0 and 1.

    """
    is_join = core.is_similarity_join(ts, query)
    if not is_join:
        query = ts

    # data conversion to np.array
    ts = core.to_np_array(ts)
    query = core.to_np_array(query)

    # validate step_size
    if not isinstance(step_size, float) or step_size > 1 or step_size < 0:
        raise ValueError('step_size should be a float between 0 and 1.')

    # validate sample_pct
    if not isinstance(sample_pct, float) or sample_pct > 1 or sample_pct < 0:
        raise ValueError('sample_pct should be a float between 0 and 1.')

    # validate random_state
    if random_state is not None:
        try:
            np.random.seed(random_state)
        except:
            raise ValueError('Invalid random_state value given.')

    if window_size < 4:
        error = "window size must be at least 4."
        raise ValueError(error)

    if window_size > len(query) / 2:
        error = "Time series is too short relative to desired window size"
        raise ValueError(error)

    # precompute some common values - profile length, query length etc.
    step_size = int(math.floor(window_size * step_size))
    profile_length = core.get_profile_length(ts, query, window_size)
    data_length = len(ts)
    exclusion_zone = int(np.ceil(window_size / 4.0))

    matrix_profile = np.zeros(profile_length)
    mp_index = np.zeros(profile_length, dtype='int')

    X = np.fft.fft(ts)
    mux, sigx = core.moving_avg_std(ts, window_size)

    dotproduct = np.zeros(profile_length)
    refine_distance = np.full(profile_length, np.inf)
    orig_index = np.arange(profile_length)

    # iterate over sampled indices and update the matrix profile
    # compute_order = compute_indices(profile_length, step_size, sample_pct)
    compute_order = np.arange(0, profile_length, step=step_size)

    for iteration, idx in enumerate(compute_order):
        subsequence = ts[idx:idx + window_size]

        # compute distance profile
        distance_profile = calc_distance_profile(X, subsequence, data_length,
            window_size, mux, sigx)
        
        # apply exclusion zone
        distance_profile = core.apply_exclusion_zone(exclusion_zone, is_join,
            window_size, data_length, idx, distance_profile)

        # find and store nearest neighbor
        if iteration == 0:
            matrix_profile = distance_profile
            mp_index[:] = idx
        else:
            update_pos = distance_profile < matrix_profile
            mp_index[update_pos] = idx
            matrix_profile[update_pos] = distance_profile[update_pos]

        idx_min = np.argmin(distance_profile)
        matrix_profile[idx] = distance_profile[idx_min]
        mp_index[idx] = idx_min
        idx_nn = mp_index[idx]

        # compute the target indices
        idx_diff = idx_nn - idx
        endidx = np.min([
            profile_length - 1,
            idx + step_size - 1,
            profile_length - idx_diff - 1
        ])
        beginidx = np.max([0, idx - step_size + 1, 2 - idx_diff])

        # compute dot product and refine distance for the idx, begin idx 
        # and end idx
        dotproduct = calc_dotproduct_idx(dotproduct, window_size, 
            matrix_profile, idx, sigx, idx_nn, mux)

        dotproduct = calc_dotproduct_end_idx(ts, dotproduct, idx, window_size,
                                             endidx, idx_nn, idx_diff)

        refine_distance = calc_refine_distance_end_idx(
            refine_distance, dotproduct, idx, endidx, mux, sigx, idx_nn,
            idx_diff, window_size)
        
        dotproduct = calc_dotproduct_begin_idx(
            ts, dotproduct, beginidx, idx, idx_diff, window_size, idx_nn)

        refine_distance = calc_refine_distance_begin_idx(
            refine_distance, dotproduct, beginidx, idx, idx_diff, idx_nn, 
            sigx, mux, window_size)

        matrix_profile, mp_index = apply_update_positions(matrix_profile, 
                                                          mp_index, 
                                                          refine_distance, 
                                                          beginidx, 
                                                          endidx, 
                                                          orig_index, idx_diff)

    return {
        'mp': matrix_profile,
        'pi': mp_index,
        'rmp': None,
        'rpi': None,
        'lmp': None,
        'lpi': None,
        'w': window_size,
        'ez': exclusion_zone,
        'join': is_join,
        'sample_pct': sample_pct,
        'metric': 'euclidean',
        'data': {
            'ts': ts,
            'query': query if is_join else None
        },
        'class': 'MatrixProfile',
        'algorithm': 'prescrimp',
    }


def scrimp_plus_plus(ts, window_size, query=None, step_size=0.25, sample_pct=0.1,
                     random_state=None, n_jobs=1):
    """SCRIMP++ is an anytime algorithm that computes the matrix profile for a 
    given time series (ts) over a given window size (m). Essentially, it allows
    for an approximate solution to be provided for quicker analysis. In the 
    case of this implementation, sample percentage is used. An approximate
    solution is given based a sample percentage from 0 to 1. The default sample
    percentage is currently 10%.

    This algorithm was created at the University of California Riverside. For
    further academic understanding, please review this paper:

    Matrix Proﬁle XI: SCRIMP++: Time Series Motif Discovery at Interactive
    Speed. Yan Zhu, Chin-Chia Michael Yeh, Zachary Zimmerman, Kaveh Kamgar
    Eamonn Keogh, ICDM 2018.

    https://www.cs.ucr.edu/~eamonn/SCRIMP_ICDM_camera_ready_updated.pdf

    Parameters
    ----------
    ts : np.ndarray
        The time series to compute the matrix profile for.
    window_size : int
        The window size.
    query : array_like
        Optionally, a query can be provided to perform a similarity join.
    step_size : float, default 0.25
        The sampling interval for the window. The paper suggest 0.25 is the
        most practical. It should be a float value between 0 and 1.
    sample_pct : float, default = 0.1 (10%)
        Number of samples to compute distances for in the MP.
    random_state : int, default None
        Set the random seed generator for reproducible results.
    n_jobs : int, Default = 1
        Number of cpu cores to use.

    Returns
    -------
    dict : profile
        A MatrixProfile data structure.

        >>> {
        >>>    'mp': The matrix profile,
        >>>    'pi': The matrix profile 1NN indices,
        >>>    'rmp': The right matrix profile,
        >>>    'rpi': The right matrix profile 1NN indices,
        >>>    'lmp': The left matrix profile,
        >>>    'lpi': The left matrix profile 1NN indices,
        >>>    'metric': The distance metric computed for the mp,
        >>>    'w': The window size used to compute the matrix profile,
        >>>    'ez': The exclusion zone used,
        >>>    'join': Flag indicating if a similarity join was computed,
        >>>    'sample_pct': Percentage of samples used in computing the MP,
        >>>    'data': {
        >>>        'ts': Time series data,
        >>>        'query': Query data if supplied
        >>>    }
        >>>    'class': "MatrixProfile"
        >>>    'algorithm': "scrimp++"
        >>> }

    Raises
    ------
    ValueError
        If window_size < 4.
        If window_size > query length / 2.
        If ts is not a list or np.array.
        If query is not a list or np.array.
        If ts or query is not one dimensional.
        If sample_pct is not between 0 and 1.

    """
    # validate random_state
    if random_state is not None:
        try:
            np.random.seed(random_state)
        except:
            raise ValueError('Invalid random_state value given.')

    ###########################
    # PreSCRIMP
    ###########################
    profile = prescrimp(ts, window_size, query=query, step_size=step_size,
        sample_pct=sample_pct, random_state=random_state, n_jobs=n_jobs)

    # data conversion to np.array
    ts = profile['data']['ts']
    query = profile['data']['query']
    if isinstance(query, type(None)):
        query = ts

    # precompute some common values - profile length, query length etc.
    step_size = int(math.floor(window_size * step_size))
    profile_length = core.get_profile_length(ts, query, window_size)
    data_length = len(ts)
    exclusion_zone = profile['ez']
    window_size = profile['w']

    # precompute some statistics on ts
    data_mu, data_sig = core.moving_avg_std(ts, window_size)

    ###########################
    # SCRIMP
    ###########################

    # randomly sort indices for compute order
    orig_index = np.arange(profile_length)
    compute_order = np.copy(orig_index[orig_index > exclusion_zone])
    #np.random.shuffle(compute_order)

    # Only refine to provided sample_pct
    sample_size = int(np.ceil(len(compute_order) * sample_pct))
    compute_order = np.random.choice(compute_order, size=sample_size, 
        replace=False)

    # initialize some values
    curlastz = np.zeros(profile_length)
    curdistance = np.zeros(profile_length)
    dist1 = np.full(profile_length, np.inf)
    dist2 = np.full(profile_length, np.inf)

    for idx in compute_order:
        # compute last z
        curlastz[idx] = np.sum(ts[0:window_size] * ts[idx:idx + window_size])
        curlastz[idx+1:] = curlastz[idx] + np.cumsum(
            (ts[window_size:data_length - idx] * ts[idx + window_size:data_length]) -\
            (ts[0:profile_length - idx - 1] * ts[idx:profile_length - 1])
        )

        # compute distances
        curdistance[idx:] = np.sqrt(np.abs(
            2 * (window_size - (curlastz[idx:profile_length + 1] -\
                window_size * (data_mu[idx:] * data_mu[0:profile_length - idx])) /\
                (data_sig[idx:] * data_sig[0:profile_length - idx]))
        ))

        dist1[0:idx - 1] = np.inf
        dist1[idx:] = curdistance[idx:]

        dist2[0:profile_length - idx] = curdistance[idx:]
        dist2[profile_length - idx + 2:] = np.inf

        loc1 = dist1 < profile['mp']
        if loc1.any():
            profile['mp'][loc1] = dist1[loc1]
            profile['pi'][loc1] = orig_index[loc1] - idx

        loc2 = dist2 < profile['mp']
        if loc2.any():
            profile['mp'][loc2] = dist2[loc2]
            profile['pi'][loc2] = orig_index[loc2] + idx


    profile['algorithm'] = 'scrimp++'
    profile['sample_pct'] = sample_pct

    return profile



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import numpy as np

from matrixprofile import core
from matrixprofile.algorithms.mpdist import mpdist_vector


def snippets(ts, snippet_size, num_snippets=2, window_size=None):
    """
    The snippets algorithm is used to summarize your time series by
    identifying N number of representative subsequences. If you want to
    identify typical patterns in your time series, then this is the algorithm
    to use.

    Parameters
    ----------
    ts : array_like
        The time series.
    snippet_size : int
        The size of snippet desired.
    num_snippets : int, Default 2
        The number of snippets you would like to find.
    window_size : int, Default (snippet_size / 2)
        The window size.

    Returns
    -------
    list : snippets
        A list of snippets as dictionary objects with the following structure.

        >>> {
        >>> 	index: the index of the snippet,
        >>> 	snippet: the snippet values,
        >>>     neighbors: the starting indices of all subsequences similar to the current snippet
        >>>     fraction: fraction of the snippet
        >>> }

    """
    ts = core.to_np_array(ts).astype('d')
    time_series_len = len(ts)
    n = len(ts)

    if not isinstance(snippet_size, int) or snippet_size < 4:
        raise ValueError('snippet_size must be an integer >= 4')

    if n < (2 * snippet_size):
        raise ValueError('Time series is too short relative to snippet length')

    if not window_size:
        window_size = int(np.floor(snippet_size / 2))

    if window_size >= snippet_size:
        raise ValueError('window_size must be smaller than snippet_size')

    # pad end of time series with zeros
    num_zeros = int(snippet_size * np.ceil(n / snippet_size) - n)
    ts = np.append(ts, np.zeros(num_zeros))

    # compute all profiles
    indices = np.arange(0, len(ts) - snippet_size, snippet_size)
    distances = []

    for j, i in enumerate(indices):
        distance = mpdist_vector(ts, ts[i:(i + snippet_size - 1)], int(window_size))
        distances.append(distance)

    distances = np.array(distances)

    # find N snippets
    snippets = []
    minis = np.inf
    total_min = None
    for n in range(num_snippets):
        minims = np.inf

        for i in range(len(indices)):
            s = np.sum(np.minimum(distances[i, :], minis))

            if minims > s:
                minims = s
                index = i

        minis = np.minimum(distances[index, :], minis)
        actual_index = indices[index]
        snippet = ts[actual_index:actual_index + snippet_size]
        snippet_distance = distances[index]
        snippets.append({
            'index': actual_index,
            'snippet': snippet,
            'distance': snippet_distance
        })

        if isinstance(total_min, type(None)):
            total_min = snippet_distance
        else:
            total_min = np.minimum(total_min, snippet_distance)

    # compute the fraction of each snippet
    for snippet in snippets:
        mask = (snippet['distance'] <= total_min)
        # create a key "neighbors" for the snippet dict,
        # and store all the time series indices for the data represented by a snippet (arr[mask])
        arr = np.arange(len(mask))
        # max_index indicates the length of a profile, which is (n-m) in the Snippets paper)
        max_index = time_series_len - snippet_size
        # since 'ts' is padded with 0 before calculate the MPdist profile
        # all parts of the profile that are out of range [0, n-m] cannot be used as neighboring snippet indices
        snippet['neighbors'] = list(filter(lambda x : x <= max_index, arr[mask]))
        # Add the last m time series indices into the neighboring snippet indices
        if max_index in snippet['neighbors']:
            last_m_indices = list(range(max_index+1, time_series_len))
            snippet['neighbors'].extend(last_m_indices)
        snippet['fraction'] = mask.sum() / (len(ts) - snippet_size)
        total_min = total_min - mask
        del snippet['distance']

    return snippets



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os
import tempfile

import numpy as np

import matrixprofile as mp


def test_disk_to_json_and_from_json_mp():
    ts = np.random.uniform(size=1024)
    w = 32

    profile = mp.algorithms.mpx(ts, w)
    out = os.path.join(tempfile.gettempdir(), 'mp.json')
    mp.io.to_disk(profile, out)

    dprofile = mp.io.from_disk(out)

    keys = set(profile.keys())
    keysb = set(dprofile.keys())

    assert(keys == keysb)

    # check values same
    for k, v in profile.items():
        if isinstance(v, np.ndarray):
            np.testing.assert_equal(v, dprofile[k])
        elif k == 'data':
            pass
        else:
            assert(v == dprofile[k])

    np.testing.assert_equal(profile['data']['ts'], dprofile['data']['ts'])
    np.testing.assert_equal(profile['data']['query'], dprofile['data']['query'])


def test_disk_to_json_and_from_json_pmp():
    ts = np.random.uniform(size=1024)

    profile = mp.algorithms.skimp(ts)
    out = os.path.join(tempfile.gettempdir(), 'pmp.json')
    mp.io.to_disk(profile, out)

    dprofile = mp.io.from_disk(out)

    keys = set(profile.keys())
    keysb = set(dprofile.keys())

    assert(keys == keysb)

    # check values same
    for k, v in profile.items():
        if isinstance(v, np.ndarray):
            np.testing.assert_equal(v, dprofile[k])
        elif k == 'data':
            pass
        else:
            assert(v == dprofile[k])

    np.testing.assert_equal(profile['data']['ts'], dprofile['data']['ts'])


def test_disk_to_mpf_and_from_mpf_mp():
    ts = np.random.uniform(size=1024)
    w = 32

    profile = mp.algorithms.mpx(ts, w)
    out = os.path.join(tempfile.gettempdir(), 'mp.mpf')
    mp.io.to_disk(profile, out, format='mpf')

    dprofile = mp.io.from_disk(out)

    keys = set(profile.keys())
    keysb = set(dprofile.keys())

    assert(keys == keysb)

    # check values same
    for k, v in profile.items():
        if isinstance(v, np.ndarray):
            np.testing.assert_equal(v, dprofile[k])
        elif k == 'data':
            pass
        else:
            assert(v == dprofile[k])

    np.testing.assert_equal(profile['data']['ts'], dprofile['data']['ts'])
    np.testing.assert_equal(profile['data']['query'], dprofile['data']['query'])


def test_disk_to_mpf_and_from_mpf_pmp():
    ts = np.random.uniform(size=1024)

    profile = mp.algorithms.skimp(ts)
    out = os.path.join(tempfile.gettempdir(), 'pmp.mpf')
    mp.io.to_disk(profile, out, format='mpf')

    dprofile = mp.io.from_disk(out)

    keys = set(profile.keys())
    keysb = set(dprofile.keys())

    assert(keys == keysb)

    # check values same
    for k, v in profile.items():
        if isinstance(v, np.ndarray):
            np.testing.assert_equal(v, dprofile[k])
        elif k == 'data':
            pass
        else:
            assert(v == dprofile[k])

    np.testing.assert_equal(profile['data']['ts'], dprofile['data']['ts'])


# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate


# handle Python 2/3 Iterable import
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import math

import warnings

from matplotlib import pyplot as plt
import numpy as np

from matrixprofile import core
from matrixprofile.algorithms.mpx import mpx
from matrixprofile.exceptions import NoSolutionPossible


def split(lower_bound, upper_bound, middle):
    """
    Helper function to split the indices for BFS.
    """
    if lower_bound == middle:
        L = None
        R = [middle + 1, upper_bound]
    elif upper_bound == middle:
        L = [lower_bound, middle - 1]
        R = None
    else:
        L = [lower_bound, middle - 1]
        R = [middle + 1, upper_bound]
    
    return (L, R)


def binary_split(n):
    """
    Create a breadth first search for indices 0..n.

    Parameters
    ----------
    n : int
        The length of indices.

    Returns
    -------
    array_like :
        The indices to iterate to perform BFS.

    """
    # having length of 1 causes infinite loop and it just doesn't really
    # make much sense.
    # just return it
    if n < 2:
        return [0,]

    index = []
    intervals = []
    
    # always begin by exploring the first integer
    index.append(0)
    
    # after exploring first integer, split interval 2:n
    intervals.append([1, n - 1])
    
    while len(intervals) > 0:
        interval = intervals.pop(0)
        lower_bound = interval[0]
        upper_bound = interval[1]
        middle = int(math.floor((lower_bound + upper_bound) / 2))
        index.append(middle)
        
        if lower_bound == upper_bound:
            continue
        else:
            L, R = split(lower_bound, upper_bound, middle)
            
            if L is not None:
                intervals.append(L)
            
            if R is not None:
                intervals.append(R)
    
    return index


def skimp(ts, windows=None, show_progress=False, cross_correlation=False,
          pmp_obj=None, sample_pct=0.1, n_jobs=1):
    """
    Computes the Pan Matrix Profile (PMP) for the given time series. When the
    time series is only passed, windows start from 8 and increase by increments
    of 2 up to length(ts) / 2. Also, the PMP is only computed using 10% of the
    windows unless sample_pct is set to a different value.

    Note
    ----
    When windows is explicitly provided, sample_pct no longer takes affect. The
    MP for all windows provided will be computed.

    Parameters
    ----------
    ts : array_like
        The time series.
    show_progress: bool, default = False
        Show the progress in percent complete in increments of 5% by printing
        it out to the console.
    cross_correlation : bool, default = False
        Return the MP values as Pearson Correlation instead of Euclidean
        distance.
    pmp_obj : dict, default = None
        Repurpose already computed window sizes with this provided PMP. It
        should be the output of a PMP algorithm such as skimp or maximum
        subsequence.
    sample_pct : float, default = 0.1 (10%)
        Number of window sizes to compute MPs for. Decimal percent between
        0 and 1.
    n_jobs : int, Default = 1
        Number of cpu cores to use.

    Returns
    -------
    dict : profile
        A Pan-MatrixProfile data structure.
        
        >>> {
        >>>     'pmp': the pan matrix profile as a 2D array,
        >>>     'pmpi': the pmp indices,
        >>>     'data': {
        >>>         'ts': time series used,
        >>>     },
        >>>     'windows': the windows used to compute the pmp,
        >>>     'sample_pct': the sample percent used,
        >>>     'metric':The distance metric computed for the pmp,
        >>>     'algorithm': the algorithm used,
        >>>     'class': PMP
        >>> }

    Raises
    ------
    ValueError :
        1. ts is not array_like.
        2. windows is not an iterable
        3. show_progress is not a boolean.
        4. cross_correlation is not a boolean.
        5. sample_pct is not between 0 and 1.

    """
    ts = core.to_np_array(ts)
    n = len(ts)
    
    # Argument validation
    if isinstance(windows, type(None)):
        start = 8
        end = int(math.floor(len(ts) / 2))
        windows = range(start, end + 1)
        
    if not isinstance(show_progress, bool):
        raise ValueError('show_progress must be a boolean!')
    
    if not isinstance(cross_correlation, bool):
        raise ValueError('cross_correlation must be a boolean!')
    
    if not isinstance(sample_pct, (int, float)) or sample_pct > 1 or sample_pct < 0:
        raise ValueError('sample_pct must be a decimal between 0 and 1')
    
    # create a breath first search index list of our window sizes
    split_index = binary_split(len(windows))
    pmp = np.full((len(split_index), n), np.inf)
    pmpi = np.full((len(split_index), n), np.nan, dtype='int')
    idx = np.full(len(split_index), -1)
    
    # compute the sample pct index
    last_index = len(split_index)
    if sample_pct < 1:
        last_index = int(np.floor(len(split_index) * sample_pct))
        last_index = np.minimum(len(split_index), last_index)
    
    pct_shown = {}

    # compute all matrix profiles for each window size
    for i in range(last_index):
        window_size = windows[split_index[i]]

        # check if we already computed this MP given a passed in PMP
        if isinstance(pmp_obj, dict):
            cw = pmp_obj.get('windows', None)
            w_idx = np.argwhere(cw == window_size)
            
            # having the window provided, we simply copy over the data instead
            # of recomputing it
            if len(w_idx) == 1:
                w_idx = w_idx[0][0]
                pmp[split_index[i], :] = pmp_obj['pmp'][w_idx, :]
                pmpi[split_index[i], :] = pmp_obj['pmpi'][w_idx, :]

                continue

        profile = mpx(ts, window_size, cross_correlation=cross_correlation,
            n_jobs=n_jobs)
        mp = profile.get('mp')
        pi = profile.get('pi')
        pmp[split_index[i], 0:len(mp)] = mp
        pmpi[split_index[i], 0:len(pi)] = pi
        
        j = split_index[i]
        while j < last_index and idx[j] != j:
            idx[j] = split_index[i]
            j = j + 1
        
        # output the progress
        if show_progress:
            pct_complete = round((i / (last_index - 1)) * 100, 2)
            int_pct = math.floor(pct_complete)
            if int_pct % 5 == 0 and int_pct not in pct_shown:
                print('{}% complete'.format(int_pct))
                pct_shown[int_pct] = 1

    metric = 'euclidean'
    if cross_correlation:
        metric = 'pearson'

    return {
        'pmp': pmp,
        'pmpi': pmpi,
        'data': {
            'ts': ts,
        },
        'windows': np.array(windows),
        'sample_pct': sample_pct,
        'metric': metric,
        'algorithm': 'skimp',
        'class': 'PMP'
    }
    

def maximum_subsequence(ts, threshold=0.95, refine_stepsize=0.05, n_jobs=1,
    include_pmp=False, lower_window=8):
    """
    Finds the maximum subsequence length based on the threshold provided. Note
    that this threshold is domain specific requiring some knowledge about the
    underyling time series in question.

    The subsequence length starts at 8 and iteratively doubles until the
    maximum correlation coefficient is no longer met. When no solution is
    possible given the threshold, a matrixprofile.exceptions.NoSolutionPossible
    exception is raised.

    Parameters
    ----------
    ts : array_like
        The time series to analyze.
    threshold : float, Default 0.95
        The correlation coefficient used as the threshold. It should be between
        0 and 1.
    refine_stepsize : float, Default 0.05
        Used in the refinement step to find a more precise upper window. It
        should be a percentage between 0.01 and 0.99.
    n_jobs : int, Default = 1
        Number of cpu cores to use.
    include_pmp : bool, default False
        Include the PanMatrixProfile for the computed windows.
    lower_window : int, default 8
        Lower bound of subsequence length that can be altered if required.
    
    Returns
    -------
    obj :
        With include_pmp=False (default)
        int : The maximum subsequence length based on the threshold provided.
        
        With include_pmp=True
        dict : A dict containing the upper window, windows and pmp.

        >>> {
        >>>     'upper_window': The upper window,
        >>>     'windows': array_like windows used to compute the pmp,
        >>>     'pmp': the pan matrix profile as a 2D array,
        >>>     'pmpi': the pmp indices,
        >>> }

    """
    windows = np.array([], dtype='int')
    pearson = np.array([], dtype='d')
    pmp = []
    pmpi = []

    ts = core.to_np_array(ts)
    n = len(ts)
    correlation_max = np.inf
    window_size = lower_window
    max_window = int(np.floor(len(ts) / 2))
    
    def resize(mp, pi, n):
        """Helper function to resize mp and pi to be aligned with the
        PMP. Also convert pearson to euclidean."""
        mp = core.pearson_to_euclidean(profile['mp'], window_size)
        infs = np.full(n - mp.shape[0], np.inf)
        nans = np.full(n - mp.shape[0], np.nan)
        mp = np.append(mp, infs)
        pi = np.append(profile['pi'], nans)

        return (mp, pi)

    # first perform a wide search by increasing window by 2 in
    # each iteration
    while window_size <= max_window:
        profile = mpx(ts, window_size, cross_correlation=True)
        mask = ~np.isinf(profile['mp'])
        correlation_max = np.max(profile['mp'][mask])

        windows = np.append(windows, window_size)
        pearson = np.append(pearson, correlation_max)

        if include_pmp:
            mp, pi = resize(profile['mp'], profile['pi'], n)
            pmp.append(mp)
            pmpi.append(pi)

        if correlation_max < threshold:
            break

        window_size = window_size * 2

    # find last window within threshold and throw away
    # computations outside of the threshold
    mask = pearson > threshold
    pearson = pearson[mask]
    windows = windows[mask]

    if len(windows) < 1:
        raise NoSolutionPossible('Given the threshold {:.2f}, no window was ' \
                                 'found. Please try increasing your ' \
                                 'threshold.')

    window_size = windows[-1]

    if include_pmp:
        pmp = np.vstack(pmp)[mask]
        pmpi = np.vstack(pmpi)[mask]

    # refine the upper u by increase by + X% increments
    test_windows = np.arange(refine_stepsize, 1, step=refine_stepsize) + 1
    test_windows = np.append(test_windows, 2)
    test_windows = np.floor(test_windows * window_size).astype('int')

    # keep windows divisible by 2
    mask = test_windows % 2 == 1
    test_windows[mask] = test_windows[mask] + 1

    for window_size in test_windows:
        profile = mpx(ts, window_size, cross_correlation=True)
        mask = ~np.isinf(profile['mp'])
        correlation_max = np.max(profile['mp'][mask])

        windows = np.append(windows, window_size)
        pearson = np.append(pearson, correlation_max)

        if include_pmp:
            mp, pi = resize(profile['mp'], profile['pi'], n)
            pmp = np.append(pmp, [mp,], axis=0)
            pmpi = np.append(pmpi, [pi,], axis=0)

        if correlation_max < threshold:
            break

    if include_pmp:
        return {
            'upper_window': window_size,
            'windows': windows,
            'pmp': pmp,
            'pmpi': pmpi
        }

    return window_size



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import numpy as np

from matrixprofile import core
from matrixprofile.algorithms.mass2 import mass2


def pmp_top_k_motifs(profile, exclusion_zone=None, k=3, max_neighbors=10, radius=3):
    """
    Find the top K number of motifs (patterns) given a pan matrix profile. By
    default the algorithm will find up to 3 motifs (k) and up to 10 of their
    neighbors with a radius of 3 * min_dist.

    Parameters
    ----------
    profile : dict
        The output from one of the pan matrix profile algorithms.
    exclusion_zone : int, Default to algorithm ez
        Desired number of values to exclude on both sides of the motif. This
        avoids trivial matches. It defaults to half of the computed window
        size. Setting the exclusion zone to 0 makes it not apply.
    k : int, Default = 3
        Desired number of motifs to find.
    max_neighbors : int, Default = 10
        The maximum number of neighbors to include for a given motif.
    radius : int, Default = 3
        The radius is used to associate a neighbor by checking if the
        neighbor's distance is less than or equal to dist * radius

    Returns
    -------
    profile : dict
        The original input obj with the addition of the "motifs" key. The
        motifs key consists of the following structure.

        A list of dicts containing motif indices and their corresponding
        neighbor indices. Note that each index is a (row, col) index
        corresponding to the pan matrix profile.
        >>> [
        >>>     {
        >>>         'motifs': [first_index, second_index],
        >>>         'neighbors': [index, index, index ...max_neighbors]
        >>>     }
        >>> ]

    """
    if not core.is_pmp_obj(profile):
        raise ValueError('Expecting PMP data structure!')

    data = profile.get('data', None)
    ts = data.get('ts', None)
    data_len = len(ts)

    pmp = profile.get('pmp', None)
    profile_len = pmp.shape[1]
    pmpi = profile.get('pmpi', None)
    windows = profile.get('windows', None)

    # make sure we are working with Euclidean distances
    tmp = None
    if core.is_pearson_array(pmp):
        tmp = core.pearson_to_euclidean(pmp, windows)
    else:
        tmp = np.copy(pmp).astype('d')

    # replace nan and infs with infinity
    tmp[core.nan_inf_indices(tmp)] = np.inf

    motifs = []
    for _ in range(k):
        min_idx = np.unravel_index(np.argmin(tmp), tmp.shape)
        min_dist = tmp[min_idx]

        # nothing else to find...
        if core.is_nan_inf(min_dist):
            break

        # create the motif pair
        min_row_idx = min_idx[0]
        min_col_idx = min_idx[1]

        # motif pairs are respective to the column of the matching row
        first_idx = np.min([min_col_idx, pmpi[min_row_idx][min_col_idx]])
        second_idx = np.max([min_col_idx, pmpi[min_row_idx][min_col_idx]])

        # compute distance profile for first appearance
        window_size = windows[min_row_idx]
        query = ts[first_idx:first_idx + window_size]
        distance_profile = mass2(ts, query)

        # extend the distance profile to be as long as the original
        infs = np.full(profile_len - len(distance_profile), np.inf)
        distance_profile = np.append(distance_profile, infs)

        # exclude already picked motifs and neighbors
        mask = core.nan_inf_indices(pmp[min_row_idx])
        distance_profile[mask] = np.inf

        # determine the exclusion zone if not set
        if not exclusion_zone:
            exclusion_zone = int(np.floor(window_size / 2))

        # apply exclusion zone for motif pair
        for j in (first_idx, second_idx):
            distance_profile = core.apply_exclusion_zone(
                exclusion_zone,
                False,
                window_size,
                data_len,
                j,
                distance_profile
            )
            tmp2 = core.apply_exclusion_zone(
                exclusion_zone,
                False,
                window_size,
                data_len,
                j,
                tmp[min_row_idx]
            )
            tmp[min_row_idx] = tmp2

        # find up to max_neighbors
        neighbors = []
        for j in range(max_neighbors):
            neighbor_idx = np.argmin(distance_profile)
            neighbor_dist = np.real(distance_profile[neighbor_idx])
            not_in_radius = not ((radius * min_dist) >= neighbor_dist)

            # no more neighbors exist based on radius
            if core.is_nan_inf(neighbor_dist) or not_in_radius:
                break

            # add neighbor and apply exclusion zone
            neighbors.append((min_row_idx, neighbor_idx))
            distance_profile = core.apply_exclusion_zone(
                exclusion_zone,
                False,
                window_size,
                data_len,
                neighbor_idx,
                distance_profile
            )
            tmp2 = core.apply_exclusion_zone(
                exclusion_zone,
                False,
                window_size,
                data_len,
                neighbor_idx,
                tmp[min_row_idx]
            )
            tmp[min_row_idx] = tmp2

        # add the motifs and neighbors
        # note that they are (row, col) indices
        motifs.append({
            'motifs': [(min_row_idx, first_idx), (min_row_idx, second_idx)],
            'neighbors': neighbors
        })

    profile['motifs'] = motifs

    return profile


def mp_top_k_motifs(profile, exclusion_zone=None, k=3, max_neighbors=10, radius=3, use_cmp=False):
    """
    Find the top K number of motifs (patterns) given a matrix profile. By
    default the algorithm will find up to 3 motifs (k) and up to 10 of their
    neighbors with a radius of 3 * min_dist using the regular matrix profile.

    Parameters
    ----------
    profile : dict
        The output from one of the matrix profile algorithms.
    exclusion_zone : int, Default to algorithm ez
        Desired number of values to exclude on both sides of the motif. This
        avoids trivial matches. It defaults to half of the computed window
        size. Setting the exclusion zone to 0 makes it not apply.
    k : int, Default = 3
        Desired number of motifs to find.
    max_neighbors : int, Default = 10
        The maximum number of neighbors to include for a given motif.
    radius : int, Default = 3
        The radius is used to associate a neighbor by checking if the
        neighbor's distance is less than or equal to dist * radius
    use_cmp : bool, Default = False
        Use the Corrected Matrix Profile to compute the motifs.

    Returns
    -------
    dict : profile
        The original input obj with the addition of the "motifs" key. The
        motifs key consists of the following structure.

        A list of dicts containing motif indices and their corresponding
        neighbor indices.

        >>> [
        >>>    {
        >>>        'motifs': [first_index, second_index],
        >>>        'neighbors': [index, index, index ...max_neighbors]
        >>>    }
        >>> ]

    """
    if not core.is_mp_obj(profile):
        raise ValueError('Expecting MP data structure!')

    window_size = profile['w']
    data = profile.get('data', None)
    if data:
        ts = data.get('ts', None)

    data_len = len(ts)
    motifs = []
    mp = np.copy(profile['mp'])
    if use_cmp:
        mp = np.copy(profile['cmp'])
    mpi = profile['pi']

    # TODO: this is based on STOMP standards when this motif finding algorithm
    # originally came out. Should we default this to 4.0 instead? That seems
    # to be the common value now per new research.
    if exclusion_zone is None:
        exclusion_zone = profile.get('ez', None)

    for i in range(k):
        min_idx = np.argmin(mp)
        min_dist = mp[min_idx]

        # we no longer have any motifs to find as all values are nan/inf
        if core.is_nan_inf(min_dist):
            break

        # create a motif pair corresponding to the first appearance and
        # second appearance
        first_idx = np.min([min_idx, mpi[min_idx]])
        second_idx = np.max([min_idx, mpi[min_idx]])

        # compute distance profile using mass2 for first appearance
        query = ts[first_idx:first_idx + window_size]
        distance_profile = mass2(ts, query)

        # exclude already picked motifs and neighbors
        mask = core.nan_inf_indices(mp)
        distance_profile[mask] = np.inf

        # apply exclusion zone for motif pair
        for j in (first_idx, second_idx):
            distance_profile = core.apply_exclusion_zone(
                exclusion_zone,
                False,
                window_size,
                data_len,
                j,
                distance_profile
            )
            mp = core.apply_exclusion_zone(
                exclusion_zone,
                False,
                window_size,
                data_len,
                j,
                mp
            )

        # find up to max_neighbors
        neighbors = []
        for j in range(max_neighbors):
            neighbor_idx = np.argmin(distance_profile)
            neighbor_dist = distance_profile[neighbor_idx]
            not_in_radius = not ((radius * min_dist) >= neighbor_dist)

            # no more neighbors exist based on radius
            if core.is_nan_inf(neighbor_dist) or not_in_radius:
                break

            # add neighbor and apply exclusion zone
            neighbors.append(neighbor_idx)
            distance_profile = core.apply_exclusion_zone(
                exclusion_zone,
                False,
                window_size,
                data_len,
                neighbor_idx,
                distance_profile
            )
            mp = core.apply_exclusion_zone(
                exclusion_zone,
                False,
                window_size,
                data_len,
                neighbor_idx,
                mp
            )

        # add motifs and neighbors to results
        motifs.append({
            'motifs': [first_idx, second_idx],
            'neighbors': neighbors
        })

    profile['motifs'] = motifs

    return profile


def top_k_motifs(profile, exclusion_zone=None, k=3, max_neighbors=10, radius=3, use_cmp=False):
    """
    Find the top K number of motifs (patterns) given a matrix profile or a pan
    matrix profile. By default the algorithm will find up to 3 motifs (k) and
    up to 10 of their neighbors with a radius of 3 * min_dist using the
    regular matrix profile. If the profile is a Matrix Profile data structure,
    you can also use a Corrected Matrix Profile to compute the motifs.

    Parameters
    ----------
    profile : dict
        The output from one of the matrix profile algorithms.
    exclusion_zone : int, Default to algorithm ez
        Desired number of values to exclude on both sides of the motif. This
        avoids trivial matches. It defaults to half of the computed window
        size. Setting the exclusion zone to 0 makes it not apply.
    k : int, Default = 3
        Desired number of motifs to find.
    max_neighbors : int, Default = 10
        The maximum number of neighbors to include for a given motif.
    radius : int, Default = 3
        The radius is used to associate a neighbor by checking if the
        neighbor's distance is less than or equal to dist * radius
    use_cmp : bool, Default = False
        Use the Corrected Matrix Profile to compute the motifs (only for
        a Matrix Profile data structure).

    Returns
    -------
    dict : profile
        The original input profile with the addition of the "motifs" key. The
        motifs key consists of the following structure.

        A list of dicts containing motif indices and their corresponding
        neighbor indices.

        >>> [
        >>>     {
        >>>         'motifs': [first_index, second_index],
        >>>         'neighbors': [index, index, index ...max_neighbors]
        >>>     }
        >>> ]

        The index is a single value when a MatrixProfile is passed in otherwise
        the index contains a row and column index for Pan-MatrixProfile.

    """
    if not core.is_mp_or_pmp_obj(profile):
        raise ValueError('Expecting MP or PMP data structure!')

    cls = profile.get('class', None)
    func = None

    if cls == 'MatrixProfile':
        func = mp_top_k_motifs
    elif cls == 'PMP':
        func = pmp_top_k_motifs
    else:
        raise ValueError('Unsupported data structure!')

    if cls == 'PMP':
        return func(
            profile,
            exclusion_zone=exclusion_zone,
            k=k,
            max_neighbors=max_neighbors,
            radius=radius
        )
        
    return func(
        profile,
        exclusion_zone=exclusion_zone,
        k=k,
        max_neighbors=max_neighbors,
        radius=radius,
        use_cmp=use_cmp
    )



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import numpy as np

from matrixprofile import core


def apply_av(profile, av="default", custom_av=None):
    """
    Utility function that returns a MatrixProfile data structure
    with a calculated annotation vector that has been applied
    to correct the matrix profile.

    Parameters
    ----------
    profile : dict
        A MatrixProfile structure.
    av : str, Default = "default"
        The type of annotation vector to apply.
    custom_av : array_like, Default = None
        Custom annotation vector (will only be applied if av is "custom").

    Returns
    -------
    dict : profile
        A MatrixProfile data structure with a calculated annotation vector
        and a corrected matrix profile.

    Raises
    ------
    ValueError
        If profile is not a MatrixProfile data structure.
        If custom_av parameter is not array-like when using a custom av.
        If av paramter is invalid.
        If lengths of annotation vector and matrix profile are different.
        If values in annotation vector are outside [0.0, 1.0].

    """
    if not core.is_mp_obj(profile):
        raise ValueError('apply_av expects profile as an MP data structure')

    temp_av = None
    av_type = None

    if av == "default":
        temp_av = make_default_av(profile['data']['ts'], profile['w'])
        av_type = av
    elif av == "complexity":
        temp_av = make_complexity_av(profile['data']['ts'], profile['w'])
        av_type = av
    elif av == "meanstd":
        temp_av = make_meanstd_av(profile['data']['ts'], profile['w'])
        av_type = av
    elif av == "clipping":
        temp_av = make_clipping_av(profile['data']['ts'], profile['w'])
        av_type = av
    elif av == "custom":
        try:
            temp_av = core.to_np_array(custom_av)
        except ValueError:
            raise ValueError('apply_av expects custom_av to be array-like')

        av_type = av
    else:
        raise ValueError("av parameter is invalid")

    if len(temp_av) != len(profile['mp']):
        raise ValueError("Lengths of annotation vector and mp are different")

    if (temp_av < 0.0).any() or (temp_av > 1.0).any():
        raise ValueError("Annotation vector values must be between 0 and 1")

    max_val = np.max(profile['mp'])
    temp_cmp = profile['mp'] + (np.ones(len(temp_av)) - temp_av) * max_val

    profile['cmp'] = temp_cmp
    profile['av'] = temp_av
    profile['av_type'] = av_type

    return profile


def make_default_av(ts, window):
    """
    Utility function that returns an annotation vector filled with 1s
    (should not change the matrix profile).

    Parameters
    ----------
    ts : array_like
        The time series.
    window : int
        The specific window size used to compute the MatrixProfile.

    Returns
    -------
    np.array : av
        An annotation vector.

    Raises
    ------
    ValueError
        If ts is not a list or np.array.
        If ts is not one-dimensional.
        If window is not an integer.

    """
    try:
        ts = core.to_np_array(ts)
    except ValueError:
        raise ValueError('make_default_av expects ts to be array-like')

    if not core.is_one_dimensional(ts):
        raise ValueError('make_default_av expects ts to be one-dimensional')

    if not isinstance(window, int):
        raise ValueError('make_default_av expects window to be an integer')

    av = np.ones(len(ts) - window + 1)

    return av


def make_complexity_av(ts, window):
    """
    Utility function that returns an annotation vector where values are based
    on the complexity estimation of the signal.

    Parameters
    ----------
    ts : array_like
        The time series.
    window : int
        The specific window size used to compute the MatrixProfile.

    Returns
    -------
    np.array : av
        An annotation vector.

    Raises
    ------
    ValueError
        If ts is not a list or np.array.
        If ts is not one-dimensional.
        If window is not an integer.

    """
    try:
        ts = core.to_np_array(ts)
    except ValueError:
        raise ValueError('make_complexity_av expects ts to be array-like')

    if not core.is_one_dimensional(ts):
        raise ValueError('make_complexity_av expects ts to be one-dimensional')

    if not isinstance(window, int):
        raise ValueError('make_complexity_av expects window to be an integer')

    av = np.zeros(len(ts) - window + 1)

    for i in range(len(av)):
        ce = np.sum(np.diff(ts[i: i + window]) ** 2)
        av[i] = np.sqrt(ce)

    max_val, min_val = np.max(av), np.min(av)
    if max_val == 0:
        av = np.zeros(len(av))
    else:
        av = (av - min_val) / max_val

    return av


def make_meanstd_av(ts, window):
    """
    Utility function that returns an annotation vector where values are set to
    1 if the standard deviation is less than the mean of standard deviation.
    Otherwise, the values are set to 0.

    Parameters
    ----------
    ts : array_like
        The time series.
    window : int
        The specific window size used to compute the MatrixProfile.

    Returns
    -------
    np.array : av
        An annotation vector.

    Raises
    ------
    ValueError
        If ts is not a list or np.array.
        If ts is not one-dimensional.
        If window is not an integer.

    """
    try:
        ts = core.to_np_array(ts)
    except ValueError:
        raise ValueError('make_meanstd_av expects ts to be array-like')

    if not core.is_one_dimensional(ts):
        raise ValueError('make_meanstd_av expects ts to be one-dimensional')

    if not isinstance(window, int):
        raise ValueError('make_meanstd_av expects window to be an integer')

    av = np.zeros(len(ts) - window + 1)

    std = core.moving_std(ts, window)
    mu = np.mean(std)
    for i in range(len(av)):
        if std[i] < mu:
            av[i] = 1

    return av


def make_clipping_av(ts, window):
    """
    Utility function that returns an annotation vector such that
    subsequences that have more clipping have less importance.

    Parameters
    ----------
    ts : array_like
        The time series.
    window : int
        The specific window size used to compute the MatrixProfile.

    Returns
    -------
    np.array : av
        An annotation vector.

    Raises
    ------
    ValueError
        If ts is not a list or np.array.
        If ts is not one-dimensional.
        If window is not an integer.

    """
    try:
        ts = core.to_np_array(ts)
    except ValueError:
        raise ValueError('make_clipping_av expects ts to be array-like')

    if not core.is_one_dimensional(ts):
        raise ValueError('make_clipping_av expects ts to be one-dimensional')

    if not isinstance(window, int):
        raise ValueError('make_clipping_av expects window to be an integer')

    av = np.zeros(len(ts) - window + 1)

    max_val, min_val = np.max(ts), np.min(ts)
    for i in range(len(av)):
        num_clip = 0.0
        for j in range(window):
            if ts[i + j] == max_val or ts[i + j] == min_val:
                num_clip += 1
        av[i] = num_clip

    min_val = np.min(av)
    av -= min_val

    max_val = np.max(av)
    if max_val == 0:
        av = np.zeros(len(av))
    else:
        av = 1 - av / max_val

    return av



from matrixprofile.algorithms.top_k_discords import top_k_discords as discords
from matrixprofile.algorithms.top_k_motifs import top_k_motifs as motifs
from matrixprofile.algorithms.snippets import snippets
from matrixprofile.algorithms.regimes import extract_regimes as regimes
from matrixprofile.algorithms.statistics import statistics
from matrixprofile.algorithms.hierarchical_clustering import (
    hierarchical_clusters
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

"""Tests for `mass_ts` package."""

import os

import pytest

import numpy as np

from matrixprofile import core


def test_is_array_like_invalid():
    assert(core.is_array_like(1) == False)
    assert(core.is_array_like('adf') == False)
    assert(core.is_array_like({'a': 1}) == False)
    assert(core.is_array_like(set([1, 2, 3])) == False)


def test_is_array_like_valid():
    assert(core.is_array_like(np.array([1])) == True)
    assert(core.is_array_like([1, ]) == True)
    assert(core.is_array_like((1, 2,)) == True)


def test_is_one_dimensional_invalid():
    a = np.array([
        [1, 2, 3],
        [2, 3, 4],
        [3, 4, 5],
        [4, 5, 6]
    ])
    assert(core.is_one_dimensional(a) == False)


def test_is_one_dimensional_valid():
    a = np.array([1, 2, 3, 4])
    assert(core.is_one_dimensional(a) == True)


def test_to_np_array_exception():
    with pytest.raises(ValueError) as excinfo:
        core.to_np_array('s')
        assert 'Unable to convert to np.ndarray!' in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        core.to_np_array(1)
        assert 'Unable to convert to np.ndarray!' in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        core.to_np_array(set([1, 2, 3]))
        assert 'Unable to convert to np.ndarray!' in str(excinfo.value)


def test_to_np_array_valid():
    actual = core.to_np_array([1, 2, 3])
    desired = np.array([1, 2, 3])
    np.testing.assert_equal(actual, desired)

    actual = core.to_np_array((1, 2, 3,))
    desired = np.array([1, 2, 3])
    np.testing.assert_equal(actual, desired)

    actual = core.to_np_array(np.array([1, 2, 3]))
    desired = np.array([1, 2, 3])
    np.testing.assert_equal(actual, desired)


def test_precheck_series_and_query_1d_valid():
    ts = [1, 2, 3, 4, 5, 6, 7, 8]
    q = [1, 2, 3, 4]

    actual_ts, actual_q = core.precheck_series_and_query_1d(ts, q)
    np.testing.assert_equal(actual_ts, np.array(ts))
    np.testing.assert_equal(actual_q, np.array(q))


def test_precheck_series_and_query_1d_invalid():
    with pytest.raises(ValueError) as excinfo:
        core.precheck_series_and_query_1d('1', [1, 2, 3])
        assert 'Invalid ts value given. Must be array_like!' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        core.precheck_series_and_query_1d([1, 2, 3], '1')
        assert 'Invalid query value given. Must be array_like!' \
            in str(excinfo.value)


def test_is_similarity_join():
    a = [1, 2, 3]
    b = [4, 5, 6, 7]

    assert(core.is_similarity_join(a, b))
    assert(not core.is_similarity_join(a, None))


def test_get_profile_length():
    a = [1, 2, 3, 4, 5, 6, 7, 8]
    b = [1, 2, 3, 4, 5]
    m = 4
    desired = 8 - 4 + 1
    actual = core.get_profile_length(a, b, m)

    assert(desired == actual)


def test_find_skip_locations():
    a = np.array([1, np.inf, np.nan, 4, 5, 6, 7, 8])
    desired = np.array([True, True, True, False, False])
    actual = core.find_skip_locations(a, 5, 4)

    np.testing.assert_equal(actual, desired)


def test_clean_nan_inf():
    a = np.array([np.nan, 1, np.inf, 2])
    desired = np.array([0, 1, 0, 2])
    actual = core.clean_nan_inf(a)

    np.testing.assert_equal(actual, desired)

    with pytest.raises(ValueError) as excinfo:
        core.clean_nan_inf(None)
        assert 'Unable to convert to np.ndarray!' in str(excinfo.value)


def test_rolling_window():
    a = np.array([1, 2, 3, 4, 5, 6])
    actual = core.rolling_window(a, 3)
    desired = np.array([
        [1, 2, 3],
        [2, 3, 4],
        [3, 4, 5],
        [4, 5, 6]
    ])

    np.testing.assert_equal(actual, desired)


def test_moving_average():
    a = np.array([1, 2, 3, 4, 5, 6])
    actual = core.moving_average(a, 3)
    desired = np.array([2., 3., 4., 5.])

    np.testing.assert_equal(actual, desired)


def test_moving_std():
    a = np.array([1, 2, 3, 4, 5, 6])
    actual = core.moving_std(a, 3)
    desired = np.array([0.81649658, 0.81649658, 0.81649658, 0.81649658])

    np.testing.assert_almost_equal(actual, desired)


def test_moving_avg_std():
    a = np.array([1, 2, 3, 4, 5, 6])
    mu, std = core.moving_avg_std(a, 3)
    mu_desired = np.array([2., 3., 4., 5.])
    std_desired = np.array([0.81649658, 0.81649658, 0.81649658, 0.81649658])

    np.testing.assert_almost_equal(mu, mu_desired)
    np.testing.assert_almost_equal(std, std_desired)


def test_fft_convolve():
    query = np.array([1, 2, 3, 4])
    ts = np.array([4, 5, 6, 1, 2, 3, 8, 9, 1, 7, 8, 15, 20])

    dp = core.fft_convolve(ts, query)
    dp_desired = np.array([36, 28, 26, 46, 68, 50, 57, 64, 99, 148])

    np.testing.assert_almost_equal(dp, dp_desired)


def test_sliding_dot_product():
    query = np.array([1, 2, 3, 4])
    ts = np.array([4, 5, 6, 1, 2, 3, 8, 9, 1, 7, 8, 15, 20])

    dp = core.sliding_dot_product(ts, query)
    dp_desired = np.array([36, 28, 26, 46, 68, 50, 57, 64, 99, 148])

    np.testing.assert_almost_equal(dp, dp_desired)


def test_generate_batch_jobs_single_job():
    profile_length = 9
    n_jobs = 1

    desired = [[0, 9]]
    actual = list(core.generate_batch_jobs(profile_length, n_jobs))

    np.testing.assert_equal(actual, desired)


def test_generate_batch_jobs_small_ts_many_jobs():
    profile_length = 9
    n_jobs = 12

    desired = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 7), (7, 8), (8, 9)]
    actual = list(core.generate_batch_jobs(profile_length, n_jobs))

    np.testing.assert_equal(actual, desired)


def test_generate_batch_jobs_multiple_jobs():
    profile_length = 9
    n_jobs = 4

    desired = [(0, 3), (3, 6), (6, 9)]
    actual = list(core.generate_batch_jobs(profile_length, n_jobs))

    np.testing.assert_equal(actual, desired)


def test_is_nan_inf():
    assert(core.is_nan_inf(np.inf) == True)
    assert(core.is_nan_inf(np.nan) == True)
    assert(core.is_nan_inf(100) == False)


def test_is_not_nan_inf():
    assert(core.is_not_nan_inf(np.inf) == False)
    assert(core.is_not_nan_inf(np.nan) == False)
    assert(core.is_not_nan_inf(100) == True)


def test_nan_inf_indices_all_inf():
    a = np.array([np.inf, np.inf, np.inf])
    desired = np.array([True, True, True])
    actual = core.nan_inf_indices(a)

    np.testing.assert_equal(desired, actual)


def test_nan_inf_indices_all_nan():
    a = np.array([np.nan, np.nan, np.nan])
    desired = np.array([True, True, True])
    actual = core.nan_inf_indices(a)

    np.testing.assert_equal(desired, actual)


def test_nan_inf_indices_some_true():
    a = np.array([1, np.inf, np.nan])
    desired = np.array([False, True, True])
    actual = core.nan_inf_indices(a)

    np.testing.assert_equal(desired, actual)


def test_nan_inf_indices_all_false():
    a = np.array([1, 1, 1])
    desired = np.array([False, False, False])
    actual = core.nan_inf_indices(a)

    np.testing.assert_equal(desired, actual)


def test_pearson_to_euclidean_1d():
    a = np.array([0.23, 0.5, 0.34, 0.67, 0.88])
    w = [4,]
    desired = np.array([2.48193473, 2, 2.29782506, 1.62480768, 0.9797959])
    actual = core.pearson_to_euclidean(a, w)

    np.testing.assert_almost_equal(desired, actual)


def test_pearson_to_euclidean_2d():
    a = np.array([[0.23, 0.5, 0.34, 0.67, 0.88],
                  [0.23, 0.5, 0.34, 0.67, 0.88]])
    w = [4, 4]
    desired = np.array([[2.48193473, 2, 2.29782506, 1.62480768, 0.9797959],
                        [2.48193473, 2, 2.29782506, 1.62480768, 0.9797959]])
    actual = core.pearson_to_euclidean(a, w)

    np.testing.assert_almost_equal(desired, actual)


def test_is_mp_obj():
    assert(True == core.is_mp_obj({'class': 'MatrixProfile'}))
    assert(False == core.is_mp_obj('s'))
    assert(False == core.is_mp_obj({}))


def test_is_pmp_obj():
    assert(True == core.is_pmp_obj({'class': 'PMP'}))
    assert(False == core.is_pmp_obj('s'))
    assert(False == core.is_pmp_obj({}))


def test_is_mp_or_pmp_obj():
    assert(True == core.is_mp_or_pmp_obj({'class': 'PMP'}))
    assert(True == core.is_mp_or_pmp_obj({'class': 'MatrixProfile'}))
    assert(False == core.is_mp_or_pmp_obj('s'))
    assert(False == core.is_mp_or_pmp_obj({}))


def test_moving_min():
    a = np.array([1, 1, 1, 2, 0, 2])
    desired = np.array([1, 0, 0])
    actual = core.moving_min(a, window=4)

    np.testing.assert_equal(desired, actual)

    a = np.array([1, 0, 1, 2, 0, 2])
    desired = np.array([0, 0, 0])
    actual = core.moving_min(a, window=4)

    np.testing.assert_equal(desired, actual)

    a = np.array([1, 1, 1, 2, 0, 2])
    desired = np.array([1, 1, 1, 0, 0])
    actual = core.moving_min(a, window=2)

    np.testing.assert_equal(desired, actual)


def test_moving_max():
    a = np.array([1, 1, 1, 2, 0, 2])
    desired = np.array([2, 2, 2])
    actual = core.moving_max(a, window=4)

    np.testing.assert_equal(desired, actual)

    a = np.array([1, 0, 1, 2, 0, 2])
    desired = np.array([2, 2, 2])
    actual = core.moving_max(a, window=4)

    np.testing.assert_equal(desired, actual)

    a = np.array([1, 1, 1, 2, 0, 2])
    desired = np.array([1, 1, 2, 2, 2])
    actual = core.moving_max(a, window=2)

    np.testing.assert_equal(desired, actual)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms.mass2 import mass2

def test_mass2():
    ts = np.array([1, 1, 1, 2, 1, 1, 4, 5])
    query = np.array([2, 1, 1, 4])
    actual = mass2(ts, query)
    desired = np.array([
        0.67640791-1.37044402e-16j,
        3.43092352+0.00000000e+00j,
        3.43092352+1.02889035e-17j,
        0.+0.00000000e+00j,
        1.85113597+1.21452707e-17j
    ])
    
    np.testing.assert_almost_equal(actual, desired)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms.top_k_motifs import top_k_motifs

def test_all_inf():
	obj = {
		'mp': np.array([np.inf, np.inf, np.inf, np.inf, np.inf]),
		'pi': np.array([0, 0, 0, 0, 0]),
		'w': 4,
		'data': {
			'ts': np.array([1, 1, 1, 1, 1, 1, 1, 1])
		},
		'class': 'MatrixProfile'
	}

	motifs = top_k_motifs(obj)
	desired = np.array([])

	np.testing.assert_equal(motifs, desired)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile import compute

import matrixprofile
MODULE_PATH = matrixprofile.__path__[0]


def test_compute_mp_exact_no_query():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    m = 32

    profile = compute(ts, windows=m)
    assert(profile['algorithm'] == 'mpx')
    assert(profile['w'] == 32)
    assert(profile['data']['query'] == None)
    assert(profile['join'] == False)
    assert(profile['sample_pct'] == 1)
    assert(profile['class'] == 'MatrixProfile')


def test_compute_mp_exact_with_query():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    query = ts[100:200]
    m = 32

    profile = compute(ts, windows=m, query=query)
    assert(profile['algorithm'] == 'mpx')
    assert(profile['w'] == 32)
    np.testing.assert_equal(profile['data']['query'], query)
    assert(profile['join'] == True)
    assert(profile['sample_pct'] == 1)
    assert(profile['class'] == 'MatrixProfile')


def test_compute_mp_approximate():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    m = 32

    profile = compute(ts, windows=m, sample_pct=0.5)
    assert(profile['algorithm'] == 'scrimp++')
    assert(profile['w'] == 32)
    assert(profile['data']['query'] == None)
    assert(profile['join'] == False)
    assert(profile['sample_pct'] == 0.5)
    assert(profile['class'] == 'MatrixProfile')


def test_compute_pmp_no_sample_pct_windows():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    windows = np.arange(8, 32)

    profile = compute(ts, windows=windows)
    assert(profile['algorithm'] == 'skimp')
    assert(profile['class'] == 'PMP')
    assert(profile['sample_pct'] == 1)
    np.testing.assert_equal(profile['windows'], windows)


def test_compute_pmp_sample_pct_windows():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    windows = np.arange(8, 32)

    profile = compute(ts, windows=windows, sample_pct=1)
    assert(profile['algorithm'] == 'skimp')
    assert(profile['class'] == 'PMP')
    assert(profile['sample_pct'] == 1)
    np.testing.assert_equal(profile['windows'], windows)


def test_compute_pmp_no_windows():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))

    profile = compute(ts)
    assert(profile['algorithm'] == 'skimp')
    assert(profile['class'] == 'PMP')

    # sample pct is ignored when windows are provided and defaults to 1
    assert(profile['sample_pct'] == 1)


def test_compute_pmp_no_windows_sample_pct():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))

    profile = compute(ts, sample_pct=0.1)
    assert(profile['algorithm'] == 'skimp')
    assert(profile['class'] == 'PMP')

    # sample pct is ignored when windows are provided and defaults to 1
    assert(profile['sample_pct'] == 0.1)


def test_compute_mp_invalid_windows():
    ts = [3., 3., 3., 3., 3., 3., 3., 3.]

    with pytest.raises(ValueError) as excinfo:
        w = 0
        compute(ts, windows=w)
        assert 'Compute requires all window sizes to be greater than 3!' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        w = 3
        compute(ts, windows=w)
        assert 'Compute requires all window sizes to be greater than 3!' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        w = [4, 0]
        compute(ts, windows=w)
        assert 'Compute requires all window sizes to be greater than 3!' \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        w = [4, 3]
        compute(ts, windows=w)
        assert 'Compute requires all window sizes to be greater than 3!' \
            in str(excinfo.value)


def test_preprocess():
    ts = np.array([np.nan, np.inf, np.inf, np.nan, np.inf, 2, 3, 2, 3, 1, 2, 3, 4, 2,
                   np.nan, np.inf, 4, 2, 3, 4, 5, 6, 7, 8, 3, 4, 2, 3, 4, 5, 6, 7, 6,
                   5, 4, 3, np.nan, np.nan, np.inf, np.nan, np.inf, np.nan])
    m = 6
    preprocessing_kwargs = {
        'window': 5,
        'impute_method': 'median',
        'impute_direction': 'backward',
        'add_noise': False
    }

    profile = compute(ts, windows=m, preprocessing_kwargs=preprocessing_kwargs)
    preprocessed_ts = profile['data']['ts']
    assert(np.any(np.isnan(preprocessed_ts)) == False)
    assert(np.any(np.isinf(preprocessed_ts)) == False)

    # if preprocessing_kwargs=None, we disable the preprocessing procedure.
    profile = compute(ts, windows=m, preprocessing_kwargs=None)
    unprocessed_ts = profile['data']['ts']
    assert(np.any(np.isnan(unprocessed_ts)) == True)
    assert(np.any(np.isinf(unprocessed_ts)) == True)

    # check if preprocessing_kwargs is None by default.
    profile = compute(ts, windows=m)
    unprocessed_ts = profile['data']['ts']
    assert(np.any(np.isnan(unprocessed_ts)) == True)
    assert(np.any(np.isinf(unprocessed_ts)) == True)

    with pytest.raises(ValueError) as excinfo:
        compute(ts, windows=m, preprocessing_kwargs=1)
        assert "The parameter 'preprocessing_kwargs' is not dict like!" \
            in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        preprocessing_kwargs = {
            'win': 5,
            'impute_dir': 'backward',
        }
        compute(ts, windows=m, preprocessing_kwargs=preprocessing_kwargs)
        assert "invalid key(s) for preprocessing_kwargs! valid key(s) should include " \
               "{'impute_direction', 'add_noise', 'impute_method', 'window'}" \
            in str(excinfo.value)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile.algorithms.regimes import extract_regimes
from matrixprofile.algorithms.mpx import mpx

import matrixprofile

MODULE_PATH = matrixprofile.__path__[0]


def test_regimes():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    w = 32
    profile = mpx(ts, w)

    # test extract 3 regimes (default)
    profile = extract_regimes(profile)
    actual = profile['regimes']
    desired = np.array([759, 423, 583])

    np.testing.assert_array_equal(actual, desired)

    # test extract 2 regimes
    profile = extract_regimes(profile, num_regimes=2)
    actual = profile['regimes']
    desired = np.array([759, 423])

    np.testing.assert_array_equal(actual, desired)


# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import numpy as np

from matrixprofile import core

def mass2(ts, query, extras=False, threshold=1e-10):
    """
    Compute the distance profile for the given query over the given time 
    series.

    Parameters
    ----------
    ts : array_like
        The time series to search.
    query : array_like
        The query.
    extras : boolean, default False
        Optionally return additional data used to compute the matrix profile.

    Returns
    -------
    np.array, dict : distance_profile
        An array of distances np.array() or dict with extras.

        With extras:
        
        >>> {
        >>>     'distance_profile': The distance profile,
        >>>     'product': The FFT product between ts and query,
        >>>     'data_mean': The moving average of the ts over len(query),
        >>>     'query_mean': The mean of the query,
        >>>     'data_std': The moving std. of the ts over len(query),
        >>>     'query_std': The std. of the query
        >>> }

    Raises
    ------
    ValueError
        If ts is not a list or np.array.
        If query is not a list or np.array.
        If ts or query is not one dimensional.

    """
    ts, query = core.precheck_series_and_query_1d(ts, query)

    n = len(ts)
    m = len(query)
    x = ts
    y = query

    meany = np.mean(y)
    sigmay = np.std(y)
    
    meanx, sigmax = core.moving_avg_std(x, m)
    meanx = np.append(np.ones([1, len(x) - len(meanx)]), meanx)    
    sigmax = np.append(np.zeros([1, len(x) - len(sigmax)]), sigmax)

    
    y = np.append(np.flip(y), np.zeros([1, n - m]))
    
    X = np.fft.fft(x)
    Y = np.fft.fft(y)
    Y.resize(X.shape)
    Z = X * Y
    z = np.fft.ifft(Z)
    
    # do not allow divide by zero
    tmp = (sigmax[m - 1:n] * sigmay)
    tmp[tmp == 0] = 1e-12

    dist = 2 * (m - (z[m - 1:n] - m * meanx[m - 1:n] * meany) / tmp)

    # fix to handle constant values
    dist[sigmax[m - 1:n] < threshold] = m
    dist[(sigmax[m - 1:n] < threshold) & (sigmay < threshold)] = 0
    dist = np.sqrt(dist)
    
    if extras:
        return {
            'distance_profile': dist,
            'product': z,
            'data_mean': meanx,
            'query_mean': meany,
            'data_std': sigmax,
            'query_std': sigmay
        }

    return dist



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

from scipy.cluster.hierarchy import linkage, inconsistent, fcluster
from scipy.cluster.hierarchy import cophenet

from matrixprofile import core
from matrixprofile.algorithms.pairwise_dist import pairwise_dist


def hierarchical_clusters(X, window_size, t, threshold=0.05, method='single', 
                          depth=2, criterion='distance', n_jobs=1):
    """
    Cluster M time series into hierarchical clusters using agglomerative
    approach. This function is more or less a convenience wrapper around 
    SciPy's scipy.cluster.hierarchy functions, but uses the MPDist algorithm
    to compute distances between each pair of time series.

    Note
    ----
    Memory usage could potentially high depending on the length of your
    time series and how many distances are computed!
    
    Parameters
    ----------
    X : array_like
        An M x N matrix where M is the time series and N is the observations at
        a given time.
    window_size : int
        The window size used to compute the MPDist.
    t : scalar
        For criteria 'inconsistent', 'distance' or 'monocrit', this is the 
        threshold to apply when forming flat clusters.
        For 'maxclust' criteria, this would be max number of clusters 
        requested.
    threshold : float, Default 0.05
        The percentile in which the MPDist is taken from. By default it is
        set to 0.05 based on empircal research results from the paper. 
        Generally, you should not change this unless you know what you are
        doing! This value must be a float greater than 0 and less than 1.
    method : str, Default single
        The linkage algorithm to use.
        Options: {single, complete, average, weighted}
    depth : int, Default 2
        A non-negative value more than 0 to specify the number of levels below
        a non-singleton cluster to allow.
    criterion : str, Default distance
        Options: {inconsistent, distance, maxclust, monocrit}
        The criterion to use in forming flat clusters.
          ``inconsistent`` :
              If a cluster node and all its
              descendants have an inconsistent value less than or equal
              to `t`, then all its leaf descendants belong to the
              same flat cluster. When no non-singleton cluster meets
              this criterion, every node is assigned to its own
              cluster. (Default)
          ``distance`` :
              Forms flat clusters so that the original
              observations in each flat cluster have no greater a
              cophenetic distance than `t`.
          ``maxclust`` :
              Finds a minimum threshold ``r`` so that
              the cophenetic distance between any two original
              observations in the same flat cluster is no more than
              ``r`` and no more than `t` flat clusters are formed.
          ``monocrit`` :
              Forms a flat cluster from a cluster node c
              with index i when ``monocrit[j] <= t``.
              For example, to threshold on the maximum mean distance
              as computed in the inconsistency matrix R with a
              threshold of 0.8 do::
                  MR = maxRstat(Z, R, 3)
                  cluster(Z, t=0.8, criterion='monocrit', monocrit=MR)
    n_jobs : int, Default 1
        The number of cpu cores used to compute the MPDist.
    
    Returns
    -------
    clusters : dict
        Clustering statistics, distances and labels.
        
        >>> {
        >>>     pairwise_distances: MPDist between pairs of time series as 
        >>>                         np.ndarray,
        >>>     linkage_matrix: clustering linkage matrix as np.ndarray,
        >>>     inconsistency_statistics: inconsistency stats as np.ndarray,
        >>>     assignments: cluster label associated with input X location as
        >>>                  np.ndarray,
        >>>     cophenet: float the cophenet statistic,
        >>>     cophenet_distances: cophenet distances between pairs of time 
        >>>                         series as np.ndarray
        >>>     class: hclusters
        >>> }
    """
    # valid SciPy clustering options to work with custom distance metric
    valid_methods = set(['single', 'complete', 'average', 'weighted'])
    valid_criterions = set([
        'inconsistent', 'distance', 'monocrit', 'maxclust'
    ])
    method = method.lower()
    criterion = criterion.lower()

    # error handling
    if not core.is_array_like(X):
        raise ValueError('X must be array like!')

    if not isinstance(t, (float, int)):
        raise ValueError('t must be a scalar (int or float)')

    if not isinstance(threshold, float) or threshold <= 0 or threshold >= 1:
        raise ValueError('threshold must be a float greater than 0 and less'\
            ' than 1')
    
    if not isinstance(depth, int) or depth < 1:
        raise ValueError('depth must be an integer greater than 0')

    if method not in valid_methods:
        opts_str = ', '.join(valid_methods)
        raise ValueError('method may only be one of: ' + opts_str)
    
    if criterion not in valid_criterions:
        opts_str = ', '.join(valid_criterions)
        raise ValueError('criterion may only be one of: ' + opts_str)

    Y = pairwise_dist(X, window_size, threshold=threshold, n_jobs=n_jobs)
    Z = linkage(Y, method=method)
    R = inconsistent(Z, d=depth)
    c, coph_dists = cophenet(Z, Y)
    T = fcluster(Z, criterion=criterion, depth=depth, R=R, t=t)
    
    return {
        'pairwise_distances': Y,
        'linkage_matrix': Z,
        'inconsistency_statistics': R,
        'assignments': T,
        'cophenet': c,
        'cophenet_distances': coph_dists,
        'class': 'hclusters'
    }



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

# Python native imports
import math
import logging

logger = logging.getLogger(__name__)

# Third-party imports
import numpy as np

# Project imports
from matrixprofile import core
from matrixprofile.preprocess import preprocess
from matrixprofile.preprocess import validate_preprocess_kwargs
from matrixprofile.algorithms.mpx import mpx
from matrixprofile.algorithms.scrimp import scrimp_plus_plus
from matrixprofile.algorithms.skimp import skimp
from matrixprofile.algorithms.skimp import maximum_subsequence


def compute(ts, windows=None, query=None, sample_pct=1, threshold=0.98,
            n_jobs=1, preprocessing_kwargs = None):
    """
    Computes the exact or approximate MatrixProfile based on the sample percent
    specified. Currently, MPX and SCRIMP++ is used for the exact and
    approximate algorithms respectively. When multiple windows are passed, the
    Pan-MatrixProfile is computed and returned.

    By default, only passing in a time series (ts), the Pan-MatrixProfile is
    computed based on the maximum upper window algorithm with a correlation
    threshold of 0.98.

    Notes
    -----
    When multiple windows are passed and the Pan-MatrixProfile is computed, the
    query is ignored!

    Parameters
    ----------
    ts : array_like
        The time series to analyze.
    windows : int, array_like
        The window(s) to compute the MatrixProfile. Note that it may be an int
        for a single matrix profile computation or an array of ints for
        computing the pan matrix profile.
    query : array_like, optional
        The query to analyze. Note that when computing the PMP the query is
        ignored!
    sample_pct : float, default 1
        A float between 0 and 1 representing how many samples to compute for
        the MP or PMP. When it is 1, the exact algorithm is used.
    threshold : float, default 0.98
        The correlation coefficient used as the threshold. It should be between
        0 and 1. This is used to compute the upper window size when no
        window(s) is given.
    n_jobs : int, default = 1
        Number of cpu cores to use.
    preprocessing_kwargs : dict, default = None
        A dictionary object to sets parameters for preprocess function.
        A valid preprocessing_kwargs should have the following structure:

        >>> {
        >>>     'window': The window size to compute the mean/median/minimum/maximum value,
        >>>     'method': A string indicating the data imputation method, which should be
        >>>               'mean', 'median', 'min' or 'max',
        >>>     'direction': A string indicating the data imputation direction, which should be
        >>>                 'forward', 'fwd', 'f', 'backward', 'bwd', 'b'. If the direction is
        >>>                 forward, we use previous data for imputation; if the direction is
        >>>                 backward, we use subsequent data for imputation.,
        >>>     'add_noise': A boolean value indicating whether noise needs to be added into the
        >>>                 time series
        >>> }

        To disable preprocessing procedure, set the preprocessing_kwargs to
        None/False/""/{}.

    Returns
    -------
    dict : profile
        The profile computed.

    """
    result = None
    multiple_windows = core.is_array_like(windows) and len(windows) > 1
    no_windows = isinstance(windows, type(None))
    has_threshold = isinstance(threshold, float)

    if no_windows and not has_threshold:
        raise ValueError('compute requires a threshold or window(s) to be set!')

    # Check to make sure all window sizes are greater than 3, return a ValueError if not.
    if (isinstance(windows, int) and windows < 4) or (multiple_windows and np.any(np.unique(windows) < 4)):
        raise ValueError('Compute requires all window sizes to be greater than 3!')

    if core.is_array_like(windows) and len(windows) == 1:
        windows = windows[0]

    # preprocess the time series
    preprocessing_kwargs = validate_preprocess_kwargs(preprocessing_kwargs)
    if preprocessing_kwargs:
        ts = preprocess(ts,
                        window=preprocessing_kwargs['window'],
                        impute_method=preprocessing_kwargs['impute_method'],
                        impute_direction=preprocessing_kwargs['impute_direction'],
                        add_noise=preprocessing_kwargs['add_noise'])

    # compute the upper window and pmp
    if no_windows and has_threshold:
        profile = maximum_subsequence(ts, threshold, include_pmp=True)

        # determine windows to be computed
        # from 8 in steps of 2 until upper w
        start = 8
        windows = range(start, profile['upper_window'] + 1)

        # compute the pmp
        result = skimp(ts, windows=windows, sample_pct=sample_pct,
                       pmp_obj=profile)

    # compute the pmp
    elif multiple_windows:
        if core.is_array_like(query):
            logger.warn('Computing PMP - query is ignored!')

        result = skimp(ts, windows=windows, sample_pct=1,
                       n_jobs=n_jobs)

    # compute exact mp
    elif sample_pct >= 1:
        result = mpx(ts, windows, query=query, n_jobs=n_jobs)

    # compute approximate mp
    else:
        result = scrimp_plus_plus(ts, windows, query=query, n_jobs=n_jobs,
                                  sample_pct=sample_pct)

    return result


# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import logging

import numpy as np

from matrixprofile import core

logger = logging.getLogger(__name__)


def _batch_compute(args):
    """
    Internal function to compute a batch of the time series in parallel.

    Parameters
    ----------
    args : tuple
        Various attributes used for computing the batch.
        (
            batch_start : int
                The starting index for this batch.
            batch_end : int
                The ending index for this batch.
            ts : array_like
                The time series to compute the matrix profile for.
            query : array_like
                The query.
            window_size : int
                The size of the window to compute the profile over.
            data_length : int
                The number of elements in the time series.
            profile_length : int
                The number of elements that will be in the final matrix
                profile.
            exclusion_zone : int
                Used to exclude trivial matches.
            is_join : bool
                Flag to indicate if an AB join or self join is occuring.
            data_mu : array_like
                The moving average over the time series for the given window
                size.
            data_sig : array_like
                The moving standard deviation over the time series for the
                given window size.
            first_product : array_like
                The first sliding dot product for the time series over index
                0 to window_size.
            skip_locs : array_like
                Indices that should be skipped for distance profile calculation
                due to a nan or inf.
        )

    Returns
    -------
    dict : profile
        The matrix profile, left and right matrix profiles and their respective
        profile indices.

        >>> {
        >>>     'mp': The matrix profile,
        >>>     'pi': The matrix profile 1NN indices,
        >>>     'rmp': The right matrix profile,
        >>>     'rpi': The right matrix profile 1NN indices,
        >>>     'lmp': The left matrix profile,
        >>>     'lpi': The left matrix profile 1NN indices,
        >>> }

    """
    batch_start, batch_end, ts, query, window_size, data_length, \
    profile_length, exclusion_zone, is_join, data_mu, data_sig, \
    first_product, skip_locs = args

     # initialize matrices
    matrix_profile = np.full(profile_length, np.inf)
    profile_index = np.full(profile_length, 0)

    left_matrix_profile = None
    right_matrix_profile = None
    left_profile_index = None
    right_profile_index = None

    if not is_join:
        left_matrix_profile = np.copy(matrix_profile)
        right_matrix_profile = np.copy(matrix_profile)
        left_profile_index = np.copy(profile_index)
        right_profile_index = np.copy(profile_index)

    # with batch 0 we do not need to recompute the dot product
    # however with other batch windows, we need the previous iterations sliding
    # dot product
    last_product = None
    if batch_start is 0:
        first_window = query[batch_start:batch_start + window_size]
        last_product = np.copy(first_product)
    else:
        first_window = query[batch_start - 1:batch_start + window_size - 1]        
        last_product = core.fft_convolve(ts, first_window)

    query_sum = np.sum(first_window)
    query_2sum = np.sum(first_window ** 2)
    query_mu, query_sig = core.moving_avg_std(first_window, window_size)

    drop_value = first_window[0]

    # only compute the distance profile for index 0 and update
    if batch_start is 0:
        distance_profile = core.distance_profile(last_product, window_size,
         data_mu, data_sig, query_mu, query_sig)

        # apply exclusion zone
        distance_profile = core.apply_exclusion_zone(exclusion_zone, is_join,
            window_size, data_length, 0, distance_profile)
           
        # update the matrix profile
        indices = (distance_profile < matrix_profile)
        matrix_profile[indices] = distance_profile[indices]
        profile_index[indices] = 0

        # update the left matrix profile
        if not is_join:
            left_matrix_profile[indices] = distance_profile[indices]
            left_profile_index[np.argwhere(indices)] = 0

        batch_start += 1

    # make sure to compute inclusively from batch start to batch end
    # otherwise there are gaps in the profile
    if batch_end < profile_length:
        batch_end += 1

    # iteratively compute distance profile and update with element-wise mins
    for i in range(batch_start, batch_end):

        # check for nan or inf and skip
        if skip_locs[i]:
            continue

        query_window = query[i:i + window_size]
        query_sum = query_sum - drop_value + query_window[-1]
        query_2sum = query_2sum - drop_value ** 2 + query_window[-1] ** 2
        query_mu = query_sum / window_size
        query_sig2 = query_2sum / window_size - query_mu ** 2
        query_sig = np.sqrt(query_sig2)
        last_product[1:] = last_product[0:data_length - window_size] \
            - ts[0:data_length - window_size] * drop_value \
            + ts[window_size:] * query_window[-1]
        last_product[0] = first_product[i]
        drop_value = query_window[0]

        distance_profile = core.distance_profile(
            last_product, window_size, data_mu, data_sig, query_mu, query_sig)

        # apply the exclusion zone        
        distance_profile = core.apply_exclusion_zone(exclusion_zone, is_join, 
            window_size, data_length, i, distance_profile)
            
        # update the matrix profile
        indices = (distance_profile < matrix_profile)
        matrix_profile[indices] = distance_profile[indices]
        profile_index[indices] = i

        # update the left and right matrix profiles
        if not is_join:
            # find differences, shift left and update
            indices = distance_profile[i:] < left_matrix_profile[i:]
            falses = np.zeros(i).astype('bool')
            indices = np.append(falses, indices)
            left_matrix_profile[indices] = distance_profile[indices]
            left_profile_index[np.argwhere(indices)] = i

            # find differences, shift right and update
            indices = distance_profile[0:i] < right_matrix_profile[0:i]
            falses = np.zeros(profile_length - i).astype('bool')
            indices = np.append(indices, falses)
            right_matrix_profile[indices] = distance_profile[indices]
            right_profile_index[np.argwhere(indices)] = i

    return {
        'mp': matrix_profile,
        'pi': profile_index,
        'rmp': right_matrix_profile,
        'rpi': right_profile_index,
        'lmp': left_matrix_profile,
        'lpi': left_profile_index,
    }


def stomp(ts, window_size, query=None, n_jobs=1):
    """
    Computes matrix profiles for a single dimensional time series using the 
    parallelized STOMP algorithm (by default). Ray or Python's multiprocessing
    library may be used. When you have initialized Ray on your machine, 
    it takes priority over using Python's multiprocessing.

    Parameters
    ----------
    ts : array_like
        The time series to compute the matrix profile for.
    window_size: int
        The size of the window to compute the matrix profile over.
    query : array_like
        Optionally, a query can be provided to perform a similarity join.
    n_jobs : int, Default = 1
        Number of cpu cores to use.

    Returns
    -------
    dict : profile
        A MatrixProfile data structure.
        
        >>> {
        >>>     'mp': The matrix profile,
        >>>     'pi': The matrix profile 1NN indices,
        >>>     'rmp': The right matrix profile,
        >>>     'rpi': The right matrix profile 1NN indices,
        >>>     'lmp': The left matrix profile,
        >>>     'lpi': The left matrix profile 1NN indices,
        >>>     'metric': The distance metric computed for the mp,
        >>>     'w': The window size used to compute the matrix profile,
        >>>     'ez': The exclusion zone used,
        >>>     'join': Flag indicating if a similarity join was computed,
        >>>     'sample_pct': Percentage of samples used in computing the MP,
        >>>     'data': {
        >>>         'ts': Time series data,
        >>>         'query': Query data if supplied
        >>>     }
        >>>     'class': "MatrixProfile"
        >>>     'algorithm': "stomp_parallel"
        >>> }

    Raises
    ------
    ValueError
        If window_size < 4.
        If window_size > query length / 2.
        If ts is not a list or np.array.
        If query is not a list or np.array.
        If ts or query is not one dimensional.

    """
    is_join = core.is_similarity_join(ts, query)
    if not is_join:
        query = ts

    # data conversion to np.array
    ts = core.to_np_array(ts)
    query = core.to_np_array(query)

    if window_size < 4:
        error = "window size must be at least 4."
        raise ValueError(error)

    if window_size > len(query) / 2:
        error = "Time series is too short relative to desired window size"
        raise ValueError(error)

    # multiprocessing or single threaded approach
    if n_jobs == 1:
        pass
    else:
        n_jobs = core.valid_n_jobs(n_jobs)

    # precompute some common values - profile length, query length etc.
    profile_length = core.get_profile_length(ts, query, window_size)
    data_length = len(ts)
    query_length = len(query)
    num_queries = query_length - window_size + 1
    exclusion_zone = int(np.ceil(window_size / 2.0))

    # do not use exclusion zone for join
    if is_join:
        exclusion_zone = 0

    # find skip locations, clean up nan and inf in the ts and query
    skip_locs = core.find_skip_locations(ts, profile_length, window_size)
    ts = core.clean_nan_inf(ts)
    query = core.clean_nan_inf(query)

    # initialize matrices
    matrix_profile = np.full(profile_length, np.inf)
    profile_index = np.full(profile_length, 0)

    # compute left and right matrix profile when similarity join does not happen
    left_matrix_profile = None
    right_matrix_profile = None
    left_profile_index = None
    right_profile_index = None

    if not is_join:
        left_matrix_profile = np.copy(matrix_profile)
        right_matrix_profile = np.copy(matrix_profile)
        left_profile_index = np.copy(profile_index)
        right_profile_index = np.copy(profile_index)

    # precompute some statistics on ts
    data_mu, data_sig = core.moving_avg_std(ts, window_size)
    first_window = query[0:window_size]
    first_product = core.fft_convolve(ts, first_window)

    batch_windows = []
    results = []
    
    # batch compute with multiprocessing
    args = []
    for start, end in core.generate_batch_jobs(num_queries, n_jobs):
        args.append((
            start, end, ts, query, window_size, data_length,
            profile_length, exclusion_zone, is_join, data_mu, data_sig,
            first_product, skip_locs
        ))
        batch_windows.append((start, end))

    # we are running single threaded stomp - no need to initialize any
    # parallel environments.
    if n_jobs == 1 or len(args) == 1:
        results.append(_batch_compute(args[0]))
    else:
        # parallelize
        with core.mp_pool()(n_jobs) as pool:
            results = pool.map(_batch_compute, args)

    # now we combine the batch results
    if len(results) == 1:
        result = results[0]
        matrix_profile = result['mp']
        profile_index = result['pi']
        left_matrix_profile = result['lmp']
        left_profile_index = result['lpi']
        right_matrix_profile = result['rmp']
        right_profile_index = result['rpi']
    else:
        for index, result in enumerate(results):
            start = batch_windows[index][0]
            end = batch_windows[index][1]        
            
            # update the matrix profile
            indices = result['mp'] < matrix_profile
            matrix_profile[indices] = result['mp'][indices]
            profile_index[indices] = result['pi'][indices]

            # update the left and right matrix profiles
            if not is_join:
                indices = result['lmp'] < left_matrix_profile
                left_matrix_profile[indices] = result['lmp'][indices]
                left_profile_index[indices] = result['lpi'][indices]

                indices = result['rmp'] < right_matrix_profile
                right_matrix_profile[indices] = result['rmp'][indices]
                right_profile_index[indices] = result['rpi'][indices]           

    return {
        'mp': matrix_profile,
        'pi': profile_index,
        'rmp': right_matrix_profile,
        'rpi': right_profile_index,
        'lmp': left_matrix_profile,
        'lpi': left_profile_index,
        'metric': 'euclidean',
        'w': window_size,
        'ez': exclusion_zone,
        'join': is_join,
        'sample_pct': 1,
        'data': {
            'ts': ts,
            'query': query
        },
        'class': "MatrixProfile",
        'algorithm': "stomp"
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import pytest

import numpy as np

from matrixprofile.datasets.datasets import load
from matrixprofile.datasets.datasets import fetch_available


def test_load_valid():
    dataset = load('motifs-discords-small')
    assert(isinstance(dataset['data'], np.ndarray) == True)
    assert('description' in dataset)
    assert('name' in dataset)
    assert('category' in dataset)


def test_load_not_found():  
    with pytest.raises(ValueError) as excinfo:
        data = load('alksdfasdf')
        assert('Could not find dataset alksdfasdf' in str(excinfo.value))


def test_fetch_available_all():
    datasets = fetch_available()
    assert(isinstance(datasets, list) == True)
    assert(len(datasets) > 0)


def test_fetch_available_category_valid():
    datasets = fetch_available(category='real')
    assert(isinstance(datasets, list) == True)
    assert(len(datasets) > 0)


def test_fetch_available_category_invalid():
    with pytest.raises(ValueError) as excinfo:
        fetch_available('alksdsfldfsd')
        assert('category alksdsfldfsd is not a valid option.' in str(excinfo.value))


# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import numpy as np

from matrixprofile import core


def empty_mp():
    """
    Utility function that provides an empty MatrixProfile data structure.

    Returns
    -------
    dict : profile
        An empty MatrixProfile data structure.

    """
    return {
        'mp': None,
        'pi': None,
        'rmp': None,
        'rpi': None,
        'lmp': None,
        'lpi': None,
        'metric': None,
        'w': None,
        'ez': None,
        'join': None,
        'data': {
            'ts': None,
            'query': None
        },
        'class': 'MatrixProfile',
        'algorithm': None
    }


def pick_mp(profile, window):
    """
    Utility function that extracts a MatrixProfile from a Pan-MatrixProfile
    placing it into the MatrixProfile data structure.

    Parameters
    ----------
    profile : dict
        A Pan-MatrixProfile data structure.
    window : int
        The specific window size used to compute the desired MatrixProfile.

    Returns
    -------
    dict : profile
        A MatrixProfile data structure.

    Raises
    ------
    ValueError
        If profile is not a Pan-MatrixProfile data structure.
        If window is not an integer.
        If desired MatrixProfile is not found based on window.

    """

    if not core.is_pmp_obj(profile):
        raise ValueError('pluck_mp expects profile as a PMP data structure!')

    if not isinstance(window, int):
        raise ValueError('pluck_mp expects window to be an int!')

    mp_profile = empty_mp()

    # find the window index
    windows = profile.get('windows')
    window_index = np.argwhere(windows == window)

    if len(window_index) < 1:
        raise RuntimeError('Unable to find window {} in the provided PMP!'.format(window))

    window_index = window_index.flatten()[0]

    window = windows[window_index]
    mp = profile['pmp'][window_index]
    n = len(mp)
    mp_profile['mp'] = mp[0:n-window+1]
    mp_profile['pi'] = profile['pmpi'][window_index][0:n-window+1]
    mp_profile['metric'] = profile['metric']
    mp_profile['data']['ts'] = profile['data']['ts']
    mp_profile['join'] = False
    mp_profile['w'] = int(window)
    mp_profile['ez'] = int(np.floor(windows[window_index] / 4))
    mp_profile['algorithm'] = 'mpx'

    return mp_profile



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto_messages.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='proto_messages.proto',
  package='mpf',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\x14proto_messages.proto\x12\x03mpf\"<\n\x0c\x44oubleMatrix\x12\x0c\n\x04rows\x18\x01 \x01(\x04\x12\x0c\n\x04\x63ols\x18\x02 \x01(\x04\x12\x10\n\x04\x64\x61ta\x18\x03 \x03(\x01\x42\x02\x10\x01\"9\n\tIntMatrix\x12\x0c\n\x04rows\x18\x01 \x01(\x04\x12\x0c\n\x04\x63ols\x18\x02 \x01(\x04\x12\x10\n\x04\x64\x61ta\x18\x03 \x03(\x03\x42\x02\x10\x01\"$\n\x08Location\x12\x0b\n\x03row\x18\x01 \x01(\x04\x12\x0b\n\x03\x63ol\x18\x02 \x01(\x04\"H\n\x05Motif\x12\x1d\n\x06motifs\x18\x01 \x03(\x0b\x32\r.mpf.Location\x12 \n\tneighbors\x18\x02 \x03(\x0b\x32\r.mpf.Location\"\xde\x01\n\rMatrixProfile\x12\x1d\n\x02mp\x18\x01 \x01(\x0b\x32\x11.mpf.DoubleMatrix\x12\x1a\n\x02pi\x18\x02 \x01(\x0b\x32\x0e.mpf.IntMatrix\x12\x1e\n\x03lmp\x18\x03 \x01(\x0b\x32\x11.mpf.DoubleMatrix\x12\x1b\n\x03lpi\x18\x04 \x01(\x0b\x32\x0e.mpf.IntMatrix\x12\x1e\n\x03rmp\x18\x05 \x01(\x0b\x32\x11.mpf.DoubleMatrix\x12\x1b\n\x03rpi\x18\x06 \x01(\x0b\x32\x0e.mpf.IntMatrix\x12\n\n\x02\x65z\x18\t \x01(\x04\x12\x0c\n\x04join\x18\n \x01(\x08\"P\n\x10PanMatrixProfile\x12\x1e\n\x03pmp\x18\x01 \x01(\x0b\x32\x11.mpf.DoubleMatrix\x12\x1c\n\x04pmpi\x18\x02 \x01(\x0b\x32\x0e.mpf.IntMatrix\"\xf8\x02\n\tMPFOutput\x12\r\n\x05klass\x18\x01 \x01(\t\x12\x11\n\talgorithm\x18\x02 \x01(\t\x12\x0e\n\x06metric\x18\x03 \x01(\t\x12\x12\n\nsample_pct\x18\x04 \x01(\x01\x12\x13\n\x07windows\x18\x05 \x03(\x04\x42\x02\x10\x01\x12\x1d\n\x02ts\x18\x06 \x01(\x0b\x32\x11.mpf.DoubleMatrix\x12 \n\x05query\x18\x07 \x01(\x0b\x32\x11.mpf.DoubleMatrix\x12\x1e\n\x02mp\x18\x08 \x01(\x0b\x32\x12.mpf.MatrixProfile\x12\"\n\x03pmp\x18\t \x01(\x0b\x32\x15.mpf.PanMatrixProfile\x12\x1a\n\x06motifs\x18\n \x03(\x0b\x32\n.mpf.Motif\x12\x1f\n\x08\x64iscords\x18\x0b \x03(\x0b\x32\r.mpf.Location\x12\x1e\n\x03\x63mp\x18\x0c \x01(\x0b\x32\x11.mpf.DoubleMatrix\x12\x1d\n\x02\x61v\x18\r \x01(\x0b\x32\x11.mpf.DoubleMatrix\x12\x0f\n\x07\x61v_type\x18\x0e \x01(\tb\x06proto3'
)




_DOUBLEMATRIX = _descriptor.Descriptor(
  name='DoubleMatrix',
  full_name='mpf.DoubleMatrix',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rows', full_name='mpf.DoubleMatrix.rows', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cols', full_name='mpf.DoubleMatrix.cols', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='mpf.DoubleMatrix.data', index=2,
      number=3, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29,
  serialized_end=89,
)


_INTMATRIX = _descriptor.Descriptor(
  name='IntMatrix',
  full_name='mpf.IntMatrix',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rows', full_name='mpf.IntMatrix.rows', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cols', full_name='mpf.IntMatrix.cols', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='mpf.IntMatrix.data', index=2,
      number=3, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=91,
  serialized_end=148,
)


_LOCATION = _descriptor.Descriptor(
  name='Location',
  full_name='mpf.Location',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='row', full_name='mpf.Location.row', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='col', full_name='mpf.Location.col', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=150,
  serialized_end=186,
)


_MOTIF = _descriptor.Descriptor(
  name='Motif',
  full_name='mpf.Motif',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='motifs', full_name='mpf.Motif.motifs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='neighbors', full_name='mpf.Motif.neighbors', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=188,
  serialized_end=260,
)


_MATRIXPROFILE = _descriptor.Descriptor(
  name='MatrixProfile',
  full_name='mpf.MatrixProfile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='mp', full_name='mpf.MatrixProfile.mp', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pi', full_name='mpf.MatrixProfile.pi', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lmp', full_name='mpf.MatrixProfile.lmp', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lpi', full_name='mpf.MatrixProfile.lpi', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rmp', full_name='mpf.MatrixProfile.rmp', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rpi', full_name='mpf.MatrixProfile.rpi', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ez', full_name='mpf.MatrixProfile.ez', index=6,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='join', full_name='mpf.MatrixProfile.join', index=7,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=263,
  serialized_end=485,
)


_PANMATRIXPROFILE = _descriptor.Descriptor(
  name='PanMatrixProfile',
  full_name='mpf.PanMatrixProfile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pmp', full_name='mpf.PanMatrixProfile.pmp', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pmpi', full_name='mpf.PanMatrixProfile.pmpi', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=487,
  serialized_end=567,
)


_MPFOUTPUT = _descriptor.Descriptor(
  name='MPFOutput',
  full_name='mpf.MPFOutput',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='klass', full_name='mpf.MPFOutput.klass', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='algorithm', full_name='mpf.MPFOutput.algorithm', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='metric', full_name='mpf.MPFOutput.metric', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sample_pct', full_name='mpf.MPFOutput.sample_pct', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='windows', full_name='mpf.MPFOutput.windows', index=4,
      number=5, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\020\001', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ts', full_name='mpf.MPFOutput.ts', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='query', full_name='mpf.MPFOutput.query', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mp', full_name='mpf.MPFOutput.mp', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pmp', full_name='mpf.MPFOutput.pmp', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='motifs', full_name='mpf.MPFOutput.motifs', index=9,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='discords', full_name='mpf.MPFOutput.discords', index=10,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cmp', full_name='mpf.MPFOutput.cmp', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='av', full_name='mpf.MPFOutput.av', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='av_type', full_name='mpf.MPFOutput.av_type', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=570,
  serialized_end=946,
)

_MOTIF.fields_by_name['motifs'].message_type = _LOCATION
_MOTIF.fields_by_name['neighbors'].message_type = _LOCATION
_MATRIXPROFILE.fields_by_name['mp'].message_type = _DOUBLEMATRIX
_MATRIXPROFILE.fields_by_name['pi'].message_type = _INTMATRIX
_MATRIXPROFILE.fields_by_name['lmp'].message_type = _DOUBLEMATRIX
_MATRIXPROFILE.fields_by_name['lpi'].message_type = _INTMATRIX
_MATRIXPROFILE.fields_by_name['rmp'].message_type = _DOUBLEMATRIX
_MATRIXPROFILE.fields_by_name['rpi'].message_type = _INTMATRIX
_PANMATRIXPROFILE.fields_by_name['pmp'].message_type = _DOUBLEMATRIX
_PANMATRIXPROFILE.fields_by_name['pmpi'].message_type = _INTMATRIX
_MPFOUTPUT.fields_by_name['ts'].message_type = _DOUBLEMATRIX
_MPFOUTPUT.fields_by_name['query'].message_type = _DOUBLEMATRIX
_MPFOUTPUT.fields_by_name['mp'].message_type = _MATRIXPROFILE
_MPFOUTPUT.fields_by_name['pmp'].message_type = _PANMATRIXPROFILE
_MPFOUTPUT.fields_by_name['motifs'].message_type = _MOTIF
_MPFOUTPUT.fields_by_name['discords'].message_type = _LOCATION
_MPFOUTPUT.fields_by_name['cmp'].message_type = _DOUBLEMATRIX
_MPFOUTPUT.fields_by_name['av'].message_type = _DOUBLEMATRIX
DESCRIPTOR.message_types_by_name['DoubleMatrix'] = _DOUBLEMATRIX
DESCRIPTOR.message_types_by_name['IntMatrix'] = _INTMATRIX
DESCRIPTOR.message_types_by_name['Location'] = _LOCATION
DESCRIPTOR.message_types_by_name['Motif'] = _MOTIF
DESCRIPTOR.message_types_by_name['MatrixProfile'] = _MATRIXPROFILE
DESCRIPTOR.message_types_by_name['PanMatrixProfile'] = _PANMATRIXPROFILE
DESCRIPTOR.message_types_by_name['MPFOutput'] = _MPFOUTPUT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DoubleMatrix = _reflection.GeneratedProtocolMessageType('DoubleMatrix', (_message.Message,), {
  'DESCRIPTOR' : _DOUBLEMATRIX,
  '__module__' : 'proto_messages_pb2'
  # @@protoc_insertion_point(class_scope:mpf.DoubleMatrix)
  })
_sym_db.RegisterMessage(DoubleMatrix)

IntMatrix = _reflection.GeneratedProtocolMessageType('IntMatrix', (_message.Message,), {
  'DESCRIPTOR' : _INTMATRIX,
  '__module__' : 'proto_messages_pb2'
  # @@protoc_insertion_point(class_scope:mpf.IntMatrix)
  })
_sym_db.RegisterMessage(IntMatrix)

Location = _reflection.GeneratedProtocolMessageType('Location', (_message.Message,), {
  'DESCRIPTOR' : _LOCATION,
  '__module__' : 'proto_messages_pb2'
  # @@protoc_insertion_point(class_scope:mpf.Location)
  })
_sym_db.RegisterMessage(Location)

Motif = _reflection.GeneratedProtocolMessageType('Motif', (_message.Message,), {
  'DESCRIPTOR' : _MOTIF,
  '__module__' : 'proto_messages_pb2'
  # @@protoc_insertion_point(class_scope:mpf.Motif)
  })
_sym_db.RegisterMessage(Motif)

MatrixProfile = _reflection.GeneratedProtocolMessageType('MatrixProfile', (_message.Message,), {
  'DESCRIPTOR' : _MATRIXPROFILE,
  '__module__' : 'proto_messages_pb2'
  # @@protoc_insertion_point(class_scope:mpf.MatrixProfile)
  })
_sym_db.RegisterMessage(MatrixProfile)

PanMatrixProfile = _reflection.GeneratedProtocolMessageType('PanMatrixProfile', (_message.Message,), {
  'DESCRIPTOR' : _PANMATRIXPROFILE,
  '__module__' : 'proto_messages_pb2'
  # @@protoc_insertion_point(class_scope:mpf.PanMatrixProfile)
  })
_sym_db.RegisterMessage(PanMatrixProfile)

MPFOutput = _reflection.GeneratedProtocolMessageType('MPFOutput', (_message.Message,), {
  'DESCRIPTOR' : _MPFOUTPUT,
  '__module__' : 'proto_messages_pb2'
  # @@protoc_insertion_point(class_scope:mpf.MPFOutput)
  })
_sym_db.RegisterMessage(MPFOutput)


_DOUBLEMATRIX.fields_by_name['data']._options = None
_INTMATRIX.fields_by_name['data']._options = None
_MPFOUTPUT.fields_by_name['windows']._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate


import numpy as np

from matrixprofile import core
from matrixprofile.io.protobuf.proto_messages_pb2 import (
    Location, Motif, MPFOutput
)


def get_matrix_attributes(matrix):
    """
    Utility function to extract the rows, cols and flattened array from a
    numpy array so it can be stored in the MPFOutput protobuf message.

    Parameters
    ----------
    matrix : np.ndarray
        The numpy array to extract the attributes from.

    Returns
    -------
    tuple :
        A tuple containing the rows, cols and flattened array.
    """
    if not core.is_array_like(matrix) or len(matrix) < 1:
        return None, None, None

    rows = matrix.shape[0]
    cols = 0
    if len(matrix.shape) > 1:
        cols = matrix.shape[1]

    return rows, cols, matrix.flatten()


def get_windows(profile):
    """
    Utility function to format the windows from a profile structure ensuring
    that the windows are in an array.

    Parameters
    ----------
    profile : dict
        The MatrixProfile or PMP profile.

    Returns
    -------
    list :
        The window(s) in a list.
    """
    windows = []

    if core.is_mp_obj(profile):
        windows.append(profile.get('w'))
    elif core.is_pmp_obj(profile):
        windows = profile.get('windows')

    return windows


def get_proto_motif(motif):
    """
    Utility function to convert a motif from a MatrixProfile or PMP structure
    ensuring that it is compatible with the MPFOutput message.

    Note
    ----
    A single dimensional motif location will only have a row index and
    a column index of 0.

    Parameters
    ----------
    motif : dict
        The motif to convert.

    Returns
    -------
    Motif :
        The motif object for MPFOutput message.
    """
    out_motif = Motif()

    for indices in motif['motifs']:
        tmp = Location()
        tmp.row = 0
        tmp.col = 0

        # handle single integer location
        if core.is_array_like(indices):
            tmp.row = indices[0]
            tmp.col = indices[1]
        else:
            tmp.row = indices

        out_motif.motifs.append(tmp)

    for neighbor in motif['neighbors']:
        tmp = Location()
        tmp.row = 0
        tmp.col = 0

        # handle single integer location
        if core.is_array_like(neighbor):
            tmp.row = neighbor[0]
            tmp.col = neighbor[1]
        else:
            tmp.row = neighbor

        out_motif.neighbors.append(tmp)

    return out_motif


def get_proto_discord(discord):
    """
    Utility function to convert a discord into the MPFOutput message
    format.

    Note
    ----
    A single dimensional discord location will only have a row index and
    a column index of 0.

    Parameters
    ----------
    discord : int or tuple
        The discord with row, col index or single index.

    Returns
    -------
    Location :
        The Location message used in the MPFOutput protobuf message.
    """
    out_discord = Location()
    out_discord.row = 0
    out_discord.col = 0

    if core.is_array_like(discord):
        out_discord.row = discord[0]
        out_discord.col = discord[1]
    else:
        out_discord.row = discord

    return out_discord


def profile_to_proto(profile):
    """
    Utility function that takes a MatrixProfile or PMP profile data structure
    and converts it to the MPFOutput protobuf message object.

    Parameters
    ----------
    profile : dict
        The profile to convert.

    Returns
    -------
    MPFOutput :
        The MPFOutput protobuf message object.
    """
    output = MPFOutput()

    # add higher level attributes that work for PMP and MP
    output.klass = profile.get('class')
    output.algorithm = profile.get('algorithm')
    output.metric = profile.get('metric')
    output.sample_pct = profile.get('sample_pct')

    # add time series data
    ts = profile.get('data').get('ts')
    query = profile.get('data').get('query')
    rows, cols, data = get_matrix_attributes(ts)
    output.ts.rows = rows
    output.ts.cols = cols
    output.ts.data.extend(data)

    # add query data
    query = profile.get('data').get('query')
    rows, cols, data = get_matrix_attributes(query)

    if rows and cols and core.is_array_like(data):
        output.query.rows = rows
        output.query.cols = cols
        output.query.data.extend(data)

    # add window(s)
    output.windows.extend(get_windows(profile))

    # add motifs
    motifs = profile.get('motifs')
    if not isinstance(motifs, type(None)):
        for motif in motifs:
            output.motifs.append(get_proto_motif(motif))

    # add discords
    discords = profile.get('discords')
    if not isinstance(discords, type(None)):
        for discord in discords:
            output.discords.append(get_proto_discord(discord))

    # add cmp
    cmp = profile.get('cmp')
    if not isinstance(cmp, type(None)):
        rows, cols, data = get_matrix_attributes(cmp)

        output.cmp.rows = rows
        output.cmp.cols = cols
        output.cmp.data.extend(data)

    # add av
    av = profile.get('av')
    if not isinstance(av, type(None)):
        rows, cols, data = get_matrix_attributes(av)

        output.av.rows = rows
        output.av.cols = cols
        output.av.data.extend(data)

    # add av_type
    av_type = profile.get('av_type')
    if not isinstance(av_type, type(None)) and len(av_type) > 0:
        output.av_type = av_type

    # add the matrix profile specific attributes
    if core.is_mp_obj(profile):
        output.mp.ez = profile.get('ez')
        output.mp.join = profile.get('join')

        # add mp
        rows, cols, data = get_matrix_attributes(profile.get('mp'))
        output.mp.mp.rows = rows
        output.mp.mp.cols = cols
        output.mp.mp.data.extend(data)

        # add pi
        rows, cols, data = get_matrix_attributes(profile.get('pi'))
        output.mp.pi.rows = rows
        output.mp.pi.cols = cols
        output.mp.pi.data.extend(data)

        # add lmp
        rows, cols, data = get_matrix_attributes(profile.get('lmp'))
        if rows and cols and core.is_array_like(data):
            output.mp.lmp.rows = rows
            output.mp.lmp.cols = cols
            output.mp.lmp.data.extend(data)

        # add lpi
        rows, cols, data = get_matrix_attributes(profile.get('lpi'))
        if rows and cols and core.is_array_like(data):
            output.mp.lpi.rows = rows
            output.mp.lpi.cols = cols
            output.mp.lpi.data.extend(data)

        # add rmp
        rows, cols, data = get_matrix_attributes(profile.get('rmp'))
        if rows and cols and core.is_array_like(data):
            output.mp.rmp.rows = rows
            output.mp.rmp.cols = cols
            output.mp.rmp.data.extend(data)

        # add rpi
        rows, cols, data = get_matrix_attributes(profile.get('rpi'))
        if rows and cols and core.is_array_like(data):
            output.mp.rpi.rows = rows
            output.mp.rpi.cols = cols
            output.mp.rpi.data.extend(data)

    # add the pan matrix profile specific attributes
    elif core.is_pmp_obj(profile):
        # add pmp
        rows, cols, data = get_matrix_attributes(profile.get('pmp'))
        output.pmp.pmp.rows = rows
        output.pmp.pmp.cols = cols
        output.pmp.pmp.data.extend(data)

        # add pmpi
        rows, cols, data = get_matrix_attributes(profile.get('pmpi'))
        output.pmp.pmpi.rows = rows
        output.pmp.pmpi.cols = cols
        output.pmp.pmpi.data.extend(data)

    else:
        raise ValueError('Expecting Pan-MatrixProfile or MatrixProfile!')

    return output


def to_mpf(profile):
    """
    Converts a given profile object into MPF binary file format.

    Parameters
    ----------
    profile : dict_like
        A MatrixProfile or Pan-MatrixProfile data structure.

    Returns
    -------
    str :
        The profile as a binary formatted string.
    """
    obj = profile_to_proto(profile)
    return obj.SerializeToString()


def from_proto_to_array(value):
    """
    Utility function to convert a protobuf array back into the correct
    dimensions.

    Parameters
    ----------
    value : array_like
        The array to transform.

    Returns
    -------
    np.ndarray :
        The transformed array.
    """
    if isinstance(value, type(None)) or len(value.data) < 1:
        return None

    shape = (value.rows, value.cols)
    out = np.array(value.data)

    if shape[1] > 0:
        out = out.reshape(shape)

    return out


def discords_from_proto(discords, is_one_dimensional=False):
    """
    Utility function to transform discord locations back to single dimension
    or multi-dimension location.

    Parameter
    ---------
    discords : array_like
        The protobuf formatted array.
    is_one_dimensional : boolean
        A flag to indicate if the original locations should be 1D.

    Returns
    -------
    np.ndarray :
        The transformed discord locations.
    """
    out = []

    for discord in discords:
        if is_one_dimensional:
            out.append(discord.row)
        else:
            out.append((discord.row, discord.col))

    return np.array(out, dtype=int)


def motifs_from_proto(motifs, is_one_dimensional=False):
    """
    Utility function to transform motif locations back to single dimension
    or multi-dimension location.

    Parameter
    ---------
    motifs : array_like
        The protobuf formatted array.
    is_one_dimensional : boolean
        A flag to indicate if the original locations should be 1D.

    Returns
    -------
    list :
        The transformed motif locations.
    """
    out = []

    for motif in motifs:
        tmp = {'motifs': [], 'neighbors': []}

        for location in motif.motifs:
            if is_one_dimensional:
                tmp['motifs'].append(location.row)
            else:
                tmp['motifs'].append((location.row, location.col))

        for neighbor in motif.neighbors:
            if is_one_dimensional:
                tmp['neighbors'].append(neighbor.row)
            else:
                tmp['neighbors'].append((neighbor.row, neighbor.col))

        out.append(tmp)

    return out


def from_mpf(profile):
    """
    Converts binary formatted MPFOutput message into a profile data structure.

    Parameters
    ----------
    profile : str
        The profile as a binary formatted MPFOutput message.

    Returns
    -------
    profile : dict_like
        A MatrixProfile or Pan-MatrixProfile data structure.
    """
    obj = MPFOutput()
    obj.ParseFromString(profile)

    out = {}
    is_one_dimensional = False

    # load in all higher level attributes
    out['class'] = obj.klass
    out['algorithm'] = obj.algorithm
    out['metric'] = obj.metric
    out['sample_pct'] = obj.sample_pct
    out['data'] = {
        'ts': from_proto_to_array(obj.ts),
        'query': from_proto_to_array(obj.query)
    }

    if obj.klass == 'MatrixProfile':
        out['mp'] = from_proto_to_array(obj.mp.mp)
        out['pi'] = from_proto_to_array(obj.mp.pi)
        out['lmp'] = from_proto_to_array(obj.mp.lmp)
        out['lpi'] = from_proto_to_array(obj.mp.lpi)
        out['rmp'] = from_proto_to_array(obj.mp.rmp)
        out['rpi'] = from_proto_to_array(obj.mp.rpi)
        out['ez'] = obj.mp.ez
        out['join'] = obj.mp.join
        out['w'] = obj.windows[0]

        is_one_dimensional = len(out['mp'].shape) == 1

    elif obj.klass == 'PMP':
        out['pmp'] = from_proto_to_array(obj.pmp.pmp)
        out['pmpi'] = from_proto_to_array(obj.pmp.pmpi)
        out['windows'] = np.array(obj.windows)

    if not isinstance(obj.discords, type(None)) and len(obj.discords) > 0:
        out['discords'] = discords_from_proto(
            obj.discords, is_one_dimensional=is_one_dimensional)

    if not isinstance(obj.motifs, type(None)) and len(obj.motifs) > 0:
        out['motifs'] = motifs_from_proto(
            obj.motifs, is_one_dimensional=is_one_dimensional)

    if not isinstance(obj.cmp, type(None)) and len(obj.cmp.data) > 0:
        out['cmp'] = from_proto_to_array(obj.cmp)

    if not isinstance(obj.av, type(None)) and len(obj.av.data) > 0:
        out['av'] = from_proto_to_array(obj.av)

    if not isinstance(obj.av_type, type(None)) and len(obj.av_type) > 0:
        out['av_type'] = obj.av_type

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

range = getattr(__builtins__, 'xrange', range)
# end of py2 compatability boilerplate

import os

import pytest

import numpy as np

from matrixprofile import utils
from matrixprofile import compute

import matrixprofile
MODULE_PATH = matrixprofile.__path__[0]

def test_empty_mp():
    keys = [
        'mp',
        'pi',
        'rmp',
        'rpi',
        'lmp',
        'lpi',
        'metric',
        'w',
        'ez',
        'join',
        'data',
        'class',
        'algorithm',
    ]

    empty = utils.empty_mp()

    for key in keys:
        assert(key in empty)
    
    assert('ts' in empty['data'])
    assert('query' in empty['data'])


def test_pick_mp():
    ts = np.loadtxt(os.path.join(MODULE_PATH, '..', 'tests', 'sampledata.txt'))
    n = len(ts)
    pmp = compute(ts)
    mp = utils.pick_mp(pmp, 32)

    assert(mp['w'] == 32)
    assert(mp['algorithm'] == 'mpx')
    assert(len(mp['mp']) == n - mp['w'] + 1)
    np.testing.assert_equal(mp['data']['ts'], ts)